"""Johnson Quantile-Parameterized Distributions."""

from __future__ import annotations

# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = [
    "FelixWick",
    "setoguchi-naoki",
]  # interface only. Cyclic boosting authors in cyclic_boosting package

from typing import Sequence

import numpy as np
import pandas as pd
from scipy.stats import logistic, norm

from skpro.distributions.base import BaseDistribution, _DelegatedDistribution


class QPD_Johnson(_DelegatedDistribution):
    """Johnson Quantile-Parameterized Distribution.

    A Johnson QPD distribution is parameterized by a symmetric-percentile triplet (SPT),
    at quantiles alpha, 0.5, and 1-alpha, respectively.

    see https://repositories.lib.utexas.edu/bitstream/handle/2152
        /63037/HADLOCK-DISSERTATION-2017.pdf
    Parameter names are as in the reference, except for the parameter lambda,
    which is renamed to kappa, as lambda is a reserved keyword in python.

    This class allows selection of the mode bounding type,
    i.e. semi-bounded, bounded, or unbounded.

    * if neither ``lower`` nor ``upper`` bound is given, the mode is unbounded
    * if only ``lower`` bound is given, the mode is semi-bounded
    * if both ``lower`` and ``upper`` bounds are given, the mode is bounded

    Parameters
    ----------
    alpha : float
        lower quantile of SPT (upper is ``1 - alpha``)
    qv_low : float or array_like[float]
        quantile function value of ``alpha``
    qv_median : float or array_like[float]
        quantile function value of quantile 0.5
    qv_high : float or array_like[float]
        quantile function value of quantile ``1 - alpha``
    lower : float, default = None
        lower bound of bounded range for QPD.
    upper : float, default = None
        upper bound of bounded range for QPD.
    base_dist: str, one of ``'normal'`` (default), ``'logistic'``
        options are ``'normal'`` (default) or ``'logistic'``

    Examples
    --------
    >>> from skpro.distributions.qpd import QPD_Johnson  # doctest: +SKIP

    >>> qpd = QPD_Johnson(
    ...         alpha=0.2,
    ...         qv_low=[1, 2],
    ...         qv_median=[3, 4],
    ...         qv_high=[5, 6],
    ...         lower=0
    ...       )  # doctest: +SKIP

    >>> qpd.mean()  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["setoguchi-naoki", "felix-wick", "fkiraly"],
        "maintainers": ["setoguchi-naoki"],
        # estimator tags
        # --------------
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "cdf", "ppf", "pdf"],
        "distr:measuretype": "continuous",
    }

    def __init__(
        self,
        alpha: float,
        qv_low: float | Sequence,
        qv_median: float | Sequence,
        qv_high: float | Sequence,
        lower: float | None = None,
        upper: float | None = None,
        base_dist: str | None = "normal",
        index=None,
        columns=None,
    ):
        self.alpha = alpha
        self.qv_low = qv_low
        self.qv_median = qv_median
        self.qv_high = qv_high
        self.lower = lower
        self.upper = upper
        self.base_dist = base_dist
        self.index = index
        self.columns = columns

        if lower is None:
            delegate_cls = QPD_U
            extra_params = {}
        elif upper is None:
            delegate_cls = QPD_S
            extra_params = {"lower": lower}
        else:
            delegate_cls = QPD_B
            extra_params = {"lower": lower, "upper": upper}

        params = {
            "alpha": alpha,
            "qv_low": qv_low,
            "qv_median": qv_median,
            "qv_high": qv_high,
            "base_dist": base_dist,
            "index": index,
            "columns": columns,
            **extra_params,
        }

        self.delegate_ = delegate_cls(**params)

        self.index = self.delegate_.index
        self.columns = self.delegate_.columns

        super().__init__(index=self.index, columns=self.columns)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": 0.2,
            "qv_median": 0.5,
            "qv_high": 0.8,
        }
        params2 = {
            "alpha": 0.1,
            "base_dist": "normal",
            "qv_low": [[-0.3], [-0.2], [-0.1]],
            "qv_median": [[-0.1], [0.0], [0.1]],
            "qv_high": [[0.2], [0.3], [0.4]],
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a"]),
        }
        params3 = {
            "alpha": 0.1,
            "base_dist": "normal",
            "qv_low": [0.1, 0.2, 0.3],
            "qv_median": [0.4, 0.5, 0.6],
            "qv_high": [0.7, 0.8, 0.9],
            "lower": 0.05,
        }
        params4 = {
            "alpha": 0.12,
            "base_dist": "logistic",
            "qv_low": [0.15, 0.1, 0.15],
            "qv_median": [0.45, 0.51, 0.54],
            "qv_high": [0.85, 0.83, 0.81],
            "lower": 0.05,
            "upper": 0.95,
        }
        return [params1, params2, params3, params4]


class QPD_S(BaseDistribution):
    """Johnson Quantile-Parameterized Distributions with semi-bounded mode.

    see https://repositories.lib.utexas.edu/bitstream/handle/2152
        /63037/HADLOCK-DISSERTATION-2017.pdf
    (Due to the Python keyword, the parameter lambda from
    this reference is named kappa below.)
    A distribution is parameterized by a symmetric-percentile triplet (SPT).

    Parameters
    ----------
    alpha : float
        lower quantile of SPT (upper is ``1 - alpha``)
    qv_low : float or array_like[float]
        quantile function value of ``alpha``
    qv_median : float or array_like[float]
        quantile function value of quantile 0.5
    qv_high : float or array_like[float]
        quantile function value of quantile ``1 - alpha``
    lower : float
        lower bound of semi-bounded range.
    base_dist: str, one of ``'normal'`` (default), ``'logistic'``
        options are ``'normal'`` (default) or ``'logistic'``

    Examples
    --------
    >>> from skpro.distributions.qpd import QPD_S  # doctest: +SKIP

    >>> qpd = QPD_S(
    ...         alpha=0.2,
    ...         qv_low=[1, 2],
    ...         qv_median=[3, 4],
    ...         qv_high=[5, 6],
    ...         lower=0
    ...       )  # doctest: +SKIP

    >>> qpd.mean()  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["setoguchi-naoki", "felix-wick", "fkiraly"],
        "maintainers": ["setoguchi-naoki"],
        # estimator tags
        # --------------
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "cdf", "ppf", "pdf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
        "broadcast_params": [
            "alpha",
            "qv_low",
            "qv_median",
            "qv_high",
            "lower",
        ],
    }

    def __init__(
        self,
        alpha: float,
        qv_low: float | Sequence,
        qv_median: float | Sequence,
        qv_high: float | Sequence,
        lower: float,
        base_dist: str | None = "normal",
        index=None,
        columns=None,
    ):
        self.alpha = alpha
        self.qv_low = qv_low
        self.qv_median = qv_median
        self.qv_high = qv_high
        self.lower = lower
        self.base_dist = base_dist
        self.index = index
        self.columns = columns

        super().__init__(index=index, columns=columns)

        # precompute parameters for methods
        phi = _resolve_phi(self.base_dist)
        self.phi = phi

        qpd_params = _prep_qpd_vars(phi=phi, mode="S", **self._bc_params)
        self._qpd_params = qpd_params

    def _ppf(self, p: np.ndarray):
        """Quantile function = percent point function = inverse cdf."""
        lower = self._bc_params["lower"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        theta = self._qpd_params["theta"]

        phi = self.phi

        in_sinh = np.arcsinh(phi.ppf(p) * delta) + np.arcsinh(n * c * delta)
        in_exp = kappa * np.sinh(in_sinh)
        ppf_arr = lower + theta * np.exp(in_exp)

        return ppf_arr

    def _pdf(self, x: np.ndarray):
        """Probability density function."""
        lower = self._bc_params["lower"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        theta = self._qpd_params["theta"]

        phi = self.phi

        # we work through the chain rule for the entire nested expression in cdf
        x_ = (x - lower) / theta
        x_der = 1 / theta

        in_arcsinh = np.log(x_) / kappa
        in_arcsinh_der = x_der / (kappa * x_)

        in_sinh = np.arcsinh(in_arcsinh) - np.arcsinh(n * c * delta)
        in_sinh_der = arcsinh_der(in_arcsinh) * in_arcsinh_der

        in_cdf = np.sinh(in_sinh) / delta
        in_cdf_der = np.cosh(in_sinh) * in_sinh_der / delta

        # cdf_arr = phi.cdf(in_cdf)
        cdf_arr_der = phi.pdf(in_cdf) * in_cdf_der

        pdf_arr = cdf_arr_der
        return pdf_arr

    def _cdf(self, x: np.ndarray):
        """Cumulative distribution function."""
        lower = self._bc_params["lower"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        theta = self._qpd_params["theta"]

        phi = self.phi

        in_arcsinh = np.log((x - lower) / theta) / kappa
        in_sinh = np.arcsinh(in_arcsinh) - np.arcsinh(n * c * delta)
        cdf_arr = phi.cdf(np.sinh(in_sinh) / delta)

        return cdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": -0.3,
            "qv_median": 0.0,
            "qv_high": 0.3,
            "lower": -0.5,
            "index": pd.RangeIndex(1),
            "columns": pd.Index(["a"]),
        }
        params2 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": [[-0.3], [-0.2], [-0.1]],
            "qv_median": [[-0.1], [0.0], [0.1]],
            "qv_high": [[0.2], [0.3], [0.4]],
            "lower": -0.5,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a"]),
        }
        return [params1, params2]


class QPD_B(BaseDistribution):
    """Johnson Quantile-Parameterized Distributions with bounded mode.

    see https://repositories.lib.utexas.edu/bitstream/handle/2152
        /63037/HADLOCK-DISSERTATION-2017.pdf
    (Due to the Python keyword, the parameter lambda from
    this reference is named kappa below).
    A distribution is parameterized by a symmetric-percentile triplet (SPT).

    Parameters
    ----------
    alpha : float or array_like[float]
        lower quantile of SPT (upper is ``1 - alpha``)
    qv_low : float or array_like[float]
        quantile function value of ``alpha``
    qv_median : float or array_like[float]
        quantile function value of quantile 0.5
    qv_high : float or array_like[float]
        quantile function value of quantile ``1 - alpha``
    lower : float or array_like[float]
        lower bound of semi-bounded range.
    base_dist: str, one of ``'normal'`` (default), ``'logistic'``
        options are ``'normal'`` (default) or ``'logistic'``

    Examples
    --------
    >>> from skpro.distributions.qpd import QPD_B  # doctest: +SKIP

    >>> qpd = QPD_B(
    ...         alpha=0.2,
    ...         qv_low=[1, 2],
    ...         qv_median=[3, 4],
    ...         qv_high=[5, 6],
    ...         lower=0,
    ...         upper=10
    ...       )  # doctest: +SKIP

    >>> qpd.mean()  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["setoguchi-naoki", "felix-wick", "fkiraly"],
        "maintainers": ["setoguchi-naoki"],
        # estimator tags
        # --------------
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "cdf", "ppf", "pdf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
        "broadcast_params": [
            "alpha",
            "qv_low",
            "qv_median",
            "qv_high",
            "lower",
            "upper",
        ],
    }

    def __init__(
        self,
        alpha: float,
        qv_low: float | Sequence,
        qv_median: float | Sequence,
        qv_high: float | Sequence,
        lower: float,
        upper: float,
        base_dist: str | None = "normal",
        index=None,
        columns=None,
    ):
        self.alpha = alpha
        self.qv_low = qv_low
        self.qv_median = qv_median
        self.qv_high = qv_high
        self.lower = lower
        self.upper = upper
        self.base_dist = base_dist
        self.index = index
        self.columns = columns

        super().__init__(index=index, columns=columns)

        # precompute parameters for methods
        phi = _resolve_phi(self.base_dist)
        self.phi = phi

        qpd_params = _prep_qpd_vars(phi=phi, mode="B", **self._bc_params)
        self._qpd_params = qpd_params

    def _ppf(self, p: np.ndarray):
        """Quantile function = percent point function = inverse cdf."""
        lower = self._bc_params["lower"]
        rnge = self._qpd_params["rnge"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        xi = self._qpd_params["xi"]

        phi = self.phi

        in_cdf = xi + kappa * np.sinh(delta * (phi.ppf(p) + n * c))
        ppf_arr = lower + rnge * phi.cdf(in_cdf)
        return ppf_arr

    def _pdf(self, x: np.ndarray):
        """Probability density function."""
        lower = self._bc_params["lower"]
        rnge = self._qpd_params["rnge"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        xi = self._qpd_params["xi"]

        phi = self.phi

        # we work through the chain rule for the entire nested expression in cdf
        x_ = (x - lower) / rnge
        x_der = 1 / rnge

        phi_ppf = phi.ppf(x_)
        # derivative of ppf at z is 1 / pdf(ppf(z))
        phi_ppf_der = x_der / phi.pdf(phi.ppf(x_))

        in_arcsinh = (phi_ppf - xi) / kappa
        in_arcsinh_der = phi_ppf_der / kappa

        in_cdf = np.arcsinh(in_arcsinh) / delta - n * c
        in_cdf_der = arcsinh_der(in_arcsinh) * in_arcsinh_der / delta

        # cdf_arr = phi.cdf(in_cdf)
        cdf_arr_der = phi.pdf(in_cdf) * in_cdf_der

        pdf_arr = cdf_arr_der
        return pdf_arr

    def _cdf(self, x: np.ndarray):
        """Cumulative distribution function."""
        lower = self._bc_params["lower"]
        rnge = self._qpd_params["rnge"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]
        c = self._qpd_params["c"]
        n = self._qpd_params["n"]
        xi = self._qpd_params["xi"]

        phi = self.phi

        phi_ppf = phi.ppf((x - lower) / rnge)
        in_cdf = np.arcsinh((phi_ppf - xi) / kappa) / delta - n * c
        cdf_arr = phi.cdf(in_cdf)

        return cdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": -0.3,
            "qv_median": 0.0,
            "qv_high": 0.3,
            "lower": -0.5,
            "upper": 0.5,
            "index": pd.RangeIndex(1),
            "columns": pd.Index(["a"]),
        }
        params2 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": [[-0.3], [-0.2], [-0.1]],
            "qv_median": [[-0.1], [0.0], [0.1]],
            "qv_high": [[0.2], [0.3], [0.4]],
            "lower": -0.5,
            "upper": 0.5,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a"]),
        }
        return [params1, params2]


class QPD_U(BaseDistribution):
    """Johnson Quantile-Parameterized Distributions with bounded mode.

    see https://repositories.lib.utexas.edu/bitstream/handle/2152
        /63037/HADLOCK-DISSERTATION-2017.pdf
    (Due to the Python keyword, the parameter lambda from
    this reference is named kappa below).
    A distribution is parameterized by a symmetric-percentile triplet (SPT).

    Parameters
    ----------
    alpha : float
        lower quantile of SPT (upper is ``1 - alpha``)
    qv_low : float or array_like[float]
        quantile function value of ``alpha``
    qv_median : float or array_like[float]
        quantile function value of quantile 0.5
    qv_high : float or array_like[float]
        quantile function value of quantile ``1 - alpha``
    base_dist: str, one of ``'normal'`` (default), ``'logistic'``
        options are ``'normal'`` (default) or ``'logistic'``

    Examples
    --------
    >>> from skpro.distributions.qpd import QPD_U  # doctest: +SKIP

    >>> qpd = QPD_U(
    ...         alpha=0.2,
    ...         qv_low=[1, 2],
    ...         qv_median=[3, 4],
    ...         qv_high=[5, 6],
    ...       )  # doctest: +SKIP

    >>> qpd.mean()  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["setoguchi-naoki", "felix-wick", "fkiraly"],
        "maintainers": ["setoguchi-naoki"],
        # estimator tags
        # --------------
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "cdf", "ppf", "pdf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
        "broadcast_params": [
            "alpha",
            "qv_low",
            "qv_median",
            "qv_high",
        ],
    }

    def __init__(
        self,
        alpha: float,
        qv_low: float | Sequence,
        qv_median: float | Sequence,
        qv_high: float | Sequence,
        base_dist: str | None = "normal",
        index=None,
        columns=None,
    ):
        self.qpd = []
        self.alpha = alpha
        self.qv_low = qv_low
        self.qv_median = qv_median
        self.qv_high = qv_high
        self.base_dist = base_dist
        self.index = index
        self.columns = columns

        super().__init__(index=index, columns=columns)

        # precompute parameters for methods
        phi = _resolve_phi(self.base_dist)
        self.phi = phi

        qpd_params = _prep_qpd_vars(phi=phi, mode="U", **self._bc_params)
        self._qpd_params = qpd_params

    def _ppf(self, p: np.ndarray):
        """Quantile function = percent point function = inverse cdf."""
        alpha = self._bc_params["alpha"]
        xi = self._qpd_params["xi"]
        gamma = self._qpd_params["gamma"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]

        phi = self.phi

        width = phi.ppf(1 - alpha)
        qs = phi.ppf(p) / width

        ppf_arr = xi + kappa * np.sinh((qs - gamma) / delta)
        return ppf_arr

    def _pdf(self, x: np.ndarray):
        """Probability density function."""
        alpha = self._bc_params["alpha"]
        xi = self._qpd_params["xi"]
        gamma = self._qpd_params["gamma"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]

        phi = self.phi

        width = phi.ppf(1 - alpha)

        qs = gamma + delta * np.arcsinh((x - xi) / kappa)
        qs_der = delta * arcsinh_der((x - xi) / kappa) / kappa

        # cdf_arr = phi.cdf(qs * width)
        pdf_arr = phi.pdf(qs * width) * qs_der
        return pdf_arr

    def _cdf(self, x: np.ndarray):
        """Cumulative distribution function."""
        alpha = self._bc_params["alpha"]
        xi = self._qpd_params["xi"]
        gamma = self._qpd_params["gamma"]
        delta = self._qpd_params["delta"]
        kappa = self._qpd_params["kappa"]

        phi = self.phi

        width = phi.ppf(1 - alpha)
        qs = gamma + delta * np.arcsinh((x - xi) / kappa)

        cdf_arr = phi.cdf(qs * width)
        return cdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": -0.3,
            "qv_median": 0.0,
            "qv_high": 0.3,
            "index": pd.RangeIndex(1),
            "columns": pd.Index(["a"]),
        }
        params2 = {
            "alpha": 0.2,
            "base_dist": "normal",
            "qv_low": [[-0.3], [-0.2], [-0.1]],
            "qv_median": [[-0.1], [0.0], [0.1]],
            "qv_high": [[0.2], [0.3], [0.4]],
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a"]),
        }
        return [params1, params2]


def _resolve_phi(phi):
    """Resolve base distribution."""
    if phi == "normal":
        return norm()
    elif phi == "logistic":
        return logistic()
    else:
        return phi


def _prep_qpd_vars(
    alpha,
    qv_low,
    qv_median,
    qv_high,
    phi,
    mode="B",
    lower=None,
    upper=None,
    **kwargs,
):
    """Prepare parameters for Johnson Quantile-Parameterized Distributions.

    Parameters
    ----------
    alpha : 2D np.array
        lower quantile of SPT (upper is ``1 - alpha``)
    qv_low : 2D np.array
        quantile function value of ``alpha``
    qv_median : 2D np.array
        quantile function value of quantile 0.5
    qv_high : 2D np.array
        quantile function value of quantile ``1 - alpha``
    lower : 2D np.array
        lower bound of range.
    upper : 2D np.array
        upper bound of range.
    phi : scipy.stats.rv_continuous
        base distribution
    mode : str
        options are ``B`` (default) or ``S``
        B = bounded mode, S = lower semi-bounded mode
    """
    c = phi.ppf(1 - alpha)

    if mode == "U":
        lower = 0

    qll = qv_low - lower
    qml = qv_median - lower
    qhl = qv_high - lower

    if mode == "B":
        rnge = upper - lower

        def tfun(x):
            return phi.ppf(x / rnge)

    elif mode == "S":
        tfun = np.log
    elif mode == "U":

        def tfun(x):
            return x

    L = tfun(qll)
    H = tfun(qhl)
    B = tfun(qml)
    HL = H - L
    BL = B - L
    HB = H - B
    LH2B = L + H - 2 * B

    HBL = np.where(BL < HB, BL, HB)

    n = np.where(LH2B > 0, 1, -1)
    n = np.where(LH2B == 0, 0, n)

    if mode in ["B", "U"]:
        xi = np.where(LH2B > 0, L, H)
        xi = np.where(LH2B == 0, B, xi)
    if mode == "S":
        theta = np.where(LH2B > 0, qll, qhl)
        theta = np.where(LH2B == 0, qml, theta)
    if mode == "U":
        theta = np.where(LH2B > 0, BL / HL, HB / HL)

    if mode in ["B", "S"]:
        in_arccosh = HL / (2 * HBL)
        delta_unn = np.arccosh(in_arccosh)
        if mode == "S":
            delta_unn = np.sinh(delta_unn)
        delta = delta_unn / c
    elif mode == "U":
        delta = 1.0 / np.arccosh(1 / (2.0 * theta))
        delta = np.where(LH2B == 0, 1, delta)

    if mode == "B":
        kappa = HL / np.sinh(2 * delta * c)
    elif mode == "S":
        kappa = HBL / (delta * c)
    elif mode == "U":
        kappa = HL / np.sinh(2.0 / delta)
        kappa = np.where(LH2B == 0, HB, kappa)

    params = {
        "c": c,
        "L": L,
        "H": H,
        "B": B,
        "n": n,
        "delta": delta,
        "kappa": kappa,
    }

    if mode == "S":
        params["theta"] = theta
    if mode == "B":
        params["rnge"] = rnge
    if mode in ["B", "U"]:
        params["xi"] = xi
    if mode == "U":
        params["gamma"] = -np.sign(LH2B)

    return params


def arcsinh_der(x):
    """Return derivative of arcsinh."""
    return 1 / np.sqrt(1 + x**2)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Machine type checkers for scitypes.

Exports
-------
check_is_mtype(obj, mtype: str, scitype: str)
    checks whether obj is mtype for scitype
    returns boolean yes/no and metadata

check_raise(obj, mtype: str, scitype:str)
    checks whether obj is mtype for scitype
    returns True if passes, otherwise raises error

mtype(obj, as_scitype: str = None)
    infer the mtype of obj, considering it as as_scitype
"""

__author__ = ["fkiraly"]

__all__ = [
    "check_is_mtype",
    "check_raise",
    "mtype",
]

from functools import lru_cache

import numpy as np

from skpro.datatypes._base import BaseDatatype
from skpro.datatypes._common import _metadata_requested, _ret
from skpro.datatypes._proba import check_dict_Proba
from skpro.datatypes._registry import AMBIGUOUS_MTYPES, SCITYPE_LIST, mtype_to_scitype


def get_check_dict(soft_deps="present"):
    """Retrieve check_dict, caches the first time it is requested.

    This is to avoid repeated, time consuming crawling in generate_check_dict,
    which would otherwise be called every time check_dict is requested.

    Parameters
    ----------
    soft_deps : str, optional - one of "present", "all"
        "present" - only checks with soft dependencies present are included
        "all" - all checks are included
    """
    if soft_deps not in ["present", "all"]:
        raise ValueError(
            "Error in get_check_dict, soft_deps argument must be 'present' or 'all', "
            f"found {soft_deps}"
        )
    check_dict = generate_check_dict(soft_deps=soft_deps)
    return check_dict.copy()


@lru_cache(maxsize=1)
def generate_check_dict(soft_deps="present"):
    """Generate check_dict using lookup."""
    from skbase.utils.dependencies import _check_estimator_deps

    from skpro.utils.retrieval import _all_classes

    classes = _all_classes("skpro.datatypes")
    classes = [x[1] for x in classes]
    classes = [x for x in classes if issubclass(x, BaseDatatype)]
    classes = [x for x in classes if not x.__name__.startswith("Base")]

    # subset only to data types with soft dependencies present
    if soft_deps == "present":
        classes = [x for x in classes if _check_estimator_deps(x, severity="none")]

    check_dict = dict()
    for cls in classes:
        k = cls()
        key = k._get_key()
        check_dict[key] = k

    # temporary while refactoring
    check_dict.update(check_dict_Proba)

    return check_dict


def _check_scitype_valid(scitype: str = None):
    """Check validity of scitype."""
    check_dict = get_check_dict()
    valid_scitypes = list({x[1] for x in check_dict.keys()})

    if not isinstance(scitype, str):
        raise TypeError(f"scitype should be a str but found {type(scitype)}")

    if scitype is not None and scitype not in valid_scitypes:
        raise TypeError(scitype + " is not a supported scitype")


def _coerce_list_of_str(obj, var_name="obj"):
    """Check whether object is string or list of string.

    Parameters
    ----------
    obj - object to check
    var_name: str, optional, default="obj" - name of input in error messages

    Returns
    -------
    list of str
        equal to obj if was a list; equal to [obj] if obj was a str
        note: if obj was a list, return is not a copy, but identical

    Raises
    ------
    TypeError if obj is not a str or list of str
    """
    if isinstance(obj, str):
        obj = [obj]
    elif isinstance(obj, list):
        if not np.all([isinstance(x, str) for x in obj]):
            raise TypeError(f"{var_name} must be a string or list of strings")
    else:
        raise TypeError(f"{var_name} must be a string or list of strings")

    return obj


def check_is_mtype(
    obj,
    mtype,
    scitype: str = None,
    return_metadata=False,
    var_name="obj",
    msg_return_dict="dict",
):
    """Check object for compliance with mtype specification, return metadata.

    Parameters
    ----------
    obj - object to check
    mtype: str or list of str, mtype to check obj as
        valid mtype strings are in datatypes.MTYPE_REGISTER (1st column)
    scitype: str, optional, scitype to check obj as; default = inferred from mtype
        if inferred from mtype, list elements of mtype need not have same scitype
        valid mtype strings are in datatypes.SCITYPE_REGISTER (1st column)
    return_metadata - bool, str, or list of str, optional, default=False
        if False, returns only "valid" return
        if True, returns all three return objects
        if str, list of str, metadata return dict is subset to keys in return_metadata
    var_name: str, optional, default="obj"
        name of input in error messages
    msg_return_dict: str, one of ``"list"`` or ``"dict"``, optional, default="dict"
        whether returned msg, if returned, is a str, dict or list

        * if ``msg_return_dict="list"``,
          returned ``msg`` is ``str`` if ``mtype`` is ``str``,
          returned ``msg`` is ``list`` of ``str`` if ``mtype`` is ``list``

        * if ``msg_return_dict="dict"``,
          returned ``msg`` is ``str`` if ``mtype`` is ``str``,
          returned ``msg`` is ``dict`` of ``str`` if ``mtype`` is ``list``.
          If ``dict``, has str in ``mtype`` as key,
          and error message for mtype as value.

    Returns
    -------
    valid: bool
        whether obj is a valid object of mtype/scitype
    msg: str or list/dict of str
        error messages if object is not valid, otherwise None,
        returned only if return_metadata is True or str, list of str

        whether ``list`` or ``dict`` type is controlled via msg_return_dict

        * if ``str``: error message for tested mtype
        * if ``list``:
          list of len(mtype) with message per mtype if list, same order as mtype
        * if ``dict``: dict with mtype as key and error message for mtype as value

    metadata: dict - metadata about obj if valid, otherwise None
        returned only if ``return_metadata`` is True or str, list of str

        Keys populated depend on (assumed, otherwise identified) scitype of obj.

        Always returned:
        * "mtype": str, mtype of obj (assumed or inferred)
        * "scitype": str, scitype of obj (assumed or inferred)

        For scitype "Table":

        * "is_univariate": bool, True iff table has one variable
        * "is_empty": bool, True iff table has no variables or no instances
        * "has_nans": bool, True iff the panel contains NaN values
        * "n_instances": int, number of instances/rows in the table
        * "n_features": int, number of variables in table
        * "feature_names": list of int or object, names of variables in table

    Raises
    ------
    TypeError if no checks defined for mtype/scitype combination
    TypeError if mtype input argument is not of expected type
    """
    mtype = _coerce_list_of_str(mtype, var_name="mtype")

    check_dict = get_check_dict()
    valid_keys = check_dict.keys()

    # we loop through individual mtypes in mtype and see whether they pass the check
    #  for each check we remember whether it passed and what it returned
    if msg_return_dict is None:
        msg_return_dict = "dict"
        msg = dict()
    elif msg_return_dict == "list":
        msg = []
    elif msg_return_dict == "dict":
        msg = dict()

    found_mtype = []
    found_scitype = []

    for m in mtype:
        if scitype is None:
            scitype_of_m = mtype_to_scitype(m)
        else:
            _check_scitype_valid(scitype)
            scitype_of_m = scitype
        key = (m, scitype_of_m)
        if (m, scitype_of_m) not in valid_keys:
            raise TypeError(f"no check defined for mtype {m}, scitype {scitype_of_m}")

        res = check_dict[key](obj, return_metadata=return_metadata, var_name=var_name)

        if _metadata_requested(return_metadata):
            check_passed = res[0]
        else:
            check_passed = res

        if check_passed:
            found_mtype.append(m)
            found_scitype.append(scitype_of_m)
            final_result = res
        elif _metadata_requested(return_metadata):
            if msg_return_dict == "list":
                msg.append(res[1])
            else:
                msg[m] = res[1]

    # there are three options on the result of check_is_mtype:
    # a. two or more mtypes are found - this is unexpected and an error with checks
    if len(found_mtype) > 1:
        raise TypeError(
            f"Error in check_is_mtype, more than one mtype identified: {found_mtype}"
        )
    # b. one mtype is found - then return that mtype
    elif len(found_mtype) == 1:
        if _metadata_requested(return_metadata):
            # add the mtype return to the metadata
            final_result[2]["mtype"] = found_mtype[0]
            final_result[2]["scitype"] = found_scitype[0]
            # final_result already has right shape and dependency on return_metadata
            return final_result
        else:
            return True
    # c. no mtype is found - then return False and all error messages if requested
    else:
        if len(msg) == 1:
            if msg_return_dict == "list":
                msg = msg[0]
            else:
                msg = list(msg.values())[0]

        return _ret(False, msg, None, return_metadata)


def check_raise(obj, mtype: str, scitype: str = None, var_name: str = "input"):
    """Check object for compliance with mtype specification, raise errors.

    Parameters
    ----------
    obj - object to check
    mtype: str or list of str, mtype to check obj as
        valid mtype strings are in datatypes.MTYPE_REGISTER (1st column)
    scitype: str, optional, scitype to check obj as; default = inferred from mtype
        if inferred from mtype, list elements of mtype need not have same scitype
        valid mtype strings are in datatypes.SCITYPE_REGISTER (1st column)
    var_name: str, optional, default="input" - name of input in error messages

    Returns
    -------
    valid: bool - True if obj complies with the specification
            same as when return argument of check_is_mtype is True
            otherwise raises an error

    Raises
    ------
    TypeError with informative message if obj does not comply
    TypeError if no checks defined for mtype/scitype combination
    ValueError if mtype input argument is not of expected type
    """
    obj_long_name_for_avoiding_linter_clash = obj
    valid, msg, _ = check_is_mtype(
        obj=obj_long_name_for_avoiding_linter_clash,
        mtype=mtype,
        scitype=scitype,
        return_metadata=[],
        var_name=var_name,
        msg_return_dict="list",
    )

    if valid:
        return True
    else:
        raise TypeError(msg)


def mtype(
    obj,
    as_scitype=None,
    exclude_mtypes=AMBIGUOUS_MTYPES,
):
    """Infer the mtype of an object considered as a specific scitype.

    Parameters
    ----------
    obj : object to infer type of - any type, should comply with some mtype spec
        if as_scitype is provided, this needs to be mtype belonging to scitype
    as_scitype : str, list of str, or None, optional, default=None
        name of scitype(s) the object "obj" is considered as, finds mtype for that
        if None (default), does not assume a specific as_scitype and tests all mtypes
            generally, as_scitype should be provided for maximum efficiency
        valid scitype type strings are in datatypes.SCITYPE_REGISTER (1st column)
    exclude_mtypes : list of str, default = AMBIGUOUS_MTYPES
        which mtypes to ignore in inferring mtype, default = ambiguous ones

    Returns
    -------
    str - the inferred mtype of "obj", a valid mtype string
            or None, if obj is None
        mtype strings with explanation are in datatypes.MTYPE_REGISTER

    Raises
    ------
    TypeError if no type can be identified, or more than one type is identified
    """
    if obj is None:
        return None

    if as_scitype is not None:
        as_scitype = _coerce_list_of_str(as_scitype, var_name="as_scitype")
        for scitype in as_scitype:
            _check_scitype_valid(scitype)

    check_dict = get_check_dict()
    m_plus_scitypes = [
        (x[0], x[1]) for x in check_dict.keys() if x[0] not in exclude_mtypes
    ]

    if as_scitype is not None:
        m_plus_scitypes = [(x[0], x[1]) for x in m_plus_scitypes if x[1] in as_scitype]

    # collects mtypes that are tested as valid for obj
    mtypes_positive = []

    # collects error messages from mtypes that are tested as invalid for obj
    mtypes_negative = dict()

    for m_plus_scitype in m_plus_scitypes:
        valid, msg, _ = check_is_mtype(
            obj,
            mtype=m_plus_scitype[0],
            scitype=m_plus_scitype[1],
            return_metadata=[],
            msg_return_dict="list",
        )
        if valid:
            mtypes_positive += [m_plus_scitype[0]]
        else:
            mtypes_negative[m_plus_scitype[0]] = msg

    if len(mtypes_positive) > 1:
        raise TypeError(
            f"Error in check_is_mtype, more than one mtype identified:"
            f" {mtypes_positive}"
        )

    if len(mtypes_positive) < 1:
        msg = ""
        for mtype, error in mtypes_negative.items():
            msg += f"{mtype}: {error}\r\n"
        msg = (
            f"No valid mtype could be identified for object of type {type(obj)}. "
            f"Errors returned are as follows, in format [mtype]: [error message] \r\n"
        ) + msg
        raise TypeError(msg)

    return mtypes_positive[0]


def check_is_scitype(
    obj,
    scitype,
    return_metadata=False,
    var_name="obj",
    exclude_mtypes=AMBIGUOUS_MTYPES,
):
    """Check object for compliance with scitype specification, return metadata.

    Parameters
    ----------
    obj - object to check
    scitype: str or list of str, scitype to check obj as
        valid mtype strings are in datatypes.SCITYPE_REGISTER
    return_metadata - bool, optional, default=False
        if False, returns only "valid" return
        if True, returns all three return objects
        if str, list of str, metadata return dict is subset to keys in return_metadata
    var_name: str, optional, default="obj" - name of input in error messages
    exclude_mtypes : list of str, default = AMBIGUOUS_MTYPES
        which mtypes to ignore in inferring mtype, default = ambiguous ones

    Returns
    -------
    valid: bool - whether obj is a valid object of mtype/scitype
    msg: dict[str, str] or None
        error messages if object is not valid, otherwise None
        keys are all mtypes tested, value for key is error message for that key
    metadata: dict - metadata about obj if valid, otherwise None
            returned only if return_metadata is True
        Fields depend on scitpe.
        Always returned:
            "mtype": str, mtype of obj (assumed or inferred)
                mtype strings with explanation are in datatypes.MTYPE_REGISTER
            "scitype": str, scitype of obj (assumed or inferred)
                scitype strings with explanation are in datatypes.SCITYPE_REGISTER
        For scitype "Series":
            "is_univariate": bool, True iff series has one variable
            "is_equally_spaced": bool, True iff series index is equally spaced
            "is_empty": bool, True iff series has no variables or no instances
            "has_nans": bool, True iff the series contains NaN values
        For scitype "Panel":
            "is_univariate": bool, True iff all series in panel have one variable
            "is_equally_spaced": bool, True iff all series indices are equally spaced
            "is_equal_length": bool, True iff all series in panel are of equal length
            "is_empty": bool, True iff one or more of the series in the panel are empty
            "is_one_series": bool, True iff there is only one series in the panel
            "has_nans": bool, True iff the panel contains NaN values
            "n_instances": int, number of instances in the panel
        For scitype "Table":
            "is_univariate": bool, True iff table has one variable
            "is_empty": bool, True iff table has no variables or no instances
            "has_nans": bool, True iff the panel contains NaN values
        For scitype "Alignment":
            currently none
    Raises
    ------
    TypeError if scitype input argument is not of expected type
    """
    scitype = _coerce_list_of_str(scitype, var_name="scitype")

    for x in scitype:
        _check_scitype_valid(x)

    check_dict = get_check_dict()
    valid_keys = check_dict.keys()

    # find all the mtype keys corresponding to the scitypes
    keys = [x for x in valid_keys if x[1] in scitype and x[0] not in exclude_mtypes]

    # storing the msg return
    msg = {}
    found_mtype = []
    found_scitype = []

    for key in keys:
        res = check_dict[key](obj, return_metadata=return_metadata, var_name=var_name)

        if _metadata_requested(return_metadata):
            check_passed = res[0]
        else:
            check_passed = res

        if check_passed:
            final_result = res
            found_mtype.append(key[0])
            found_scitype.append(key[1])
        elif _metadata_requested(return_metadata):
            msg[key[0]] = res[1]

    # there are three options on the result of check_is_mtype:
    # a. two or more mtypes are found - this is unexpected and an error with checks
    if len(found_mtype) > 1:
        raise TypeError(
            f"Error in check_is_mtype, more than one mtype identified: {found_mtype}"
        )
    # b. one mtype is found - then return that mtype
    elif len(found_mtype) == 1:
        if _metadata_requested(return_metadata):
            # add the mtype return to the metadata
            final_result[2]["mtype"] = found_mtype[0]
            # add the scitype return to the metadata
            final_result[2]["scitype"] = found_scitype[0]
            # final_result already has right shape and dependency on return_metadata
            return final_result
        else:
            return True
    # c. no mtype is found - then return False and all error messages if requested
    else:
        return _ret(False, msg, None, return_metadata)


def check_is_error_msg(msg, var_name="obj", allowed_msg=None, raise_exception=False):
    """Format and possibly raise error message from check_is_mtype or check_is_scitype.

    Parameters
    ----------
    msg: dict[str, str]
        error message from check_is_scitype, or from check_is_mtype with dict return
    var_name: str, optional, default="obj"
        name of input in error messages
    allowed_msg: str, optional, default=None
        message component detailing allowed mtypes or scitype combinations
    raise_exception: bool or Exception, optional, default=False
        whether to raise exception or return error message
        if False, returns formatted error message
        if True, raises TypeError with formatted error message
        if Exception, raises that Exception with formatted error message

    Returns
    -------
    str - formatted error message
    """
    msg_invalid_input = (
        f"{var_name} must be in an skpro compatible format. {allowed_msg}"
        f"If you think the data is already in an skpro supported input format, "
        f"run skpro.datatypes.check_raise(data, mtype) to diagnose the error, "
        f"where mtype is the string of the type specification you want. "
        f"Error message for checked mtypes, in format [mtype: message], as follows:"
    )
    for mtype, err in msg.items():
        msg_invalid_input += f" [{mtype}: {err}] "

    if raise_exception is True:
        raise TypeError(msg_invalid_input)
    elif raise_exception is False:
        return msg_invalid_input
    else:
        raise raise_exception(msg_invalid_input)


def scitype(obj, candidate_scitypes=SCITYPE_LIST, exclude_mtypes=AMBIGUOUS_MTYPES):
    """Infer the scitype of an object.

    Parameters
    ----------
    obj : object to infer type of - any type, should comply with some mtype spec
        if as_scitype is provided, this needs to be mtype belonging to scitype
    candidate_scitypes: str or list of str, scitypes to pick from
        valid scitype strings are in datatypes.SCITYPE_REGISTER
    exclude_mtypes : list of str, default = AMBIGUOUS_MTYPES
        which mtypes to ignore in inferring mtype, default = ambiguous ones
        valid mtype strings are in datatypes.MTYPE_REGISTER

    Returns
    -------
    str - the inferred sciype of "obj", a valid scitype string
            or None, if obj is None
        scitype strings with explanation are in datatypes.SCITYPE_REGISTER

    Raises
    ------
    TypeError if no type can be identified, or more than one type is identified
    """
    candidate_scitypes = _coerce_list_of_str(
        candidate_scitypes, var_name="candidate_scitypes"
    )

    valid_scitypes = []

    for scitype in candidate_scitypes:
        valid = check_is_scitype(
            obj,
            scitype=scitype,
            return_metadata=False,
            exclude_mtypes=exclude_mtypes,
        )
        if valid:
            valid_scitypes += [scitype]

    if len(valid_scitypes) > 1:
        raise TypeError(
            "Error in function scitype, more than one valid scitype identified:"
            f"{valid_scitypes}"
        )
    if len(valid_scitypes) == 0:
        raise TypeError(
            "Error in function scitype, no valid scitype could be identified."
        )

    return valid_scitypes[0]



"""Metrics for probabilistic supervised regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__author__ = ["fkiraly", "euanenticott-shell"]

__all__ = [
    "CRPS",
    "AUCalibration",
    "ConstraintViolation",
    "EmpiricalCoverage",
    "IntervalWidth",
    "LogLoss",
    "LinearizedLogLoss",
    "PinballLoss",
    "SquaredDistrLoss",
    # survival metrics
    "ConcordanceHarrell",
    "SPLL",
]

from skpro.metrics._classes import (
    CRPS,
    AUCalibration,
    ConstraintViolation,
    EmpiricalCoverage,
    IntervalWidth,
    LinearizedLogLoss,
    LogLoss,
    PinballLoss,
    SquaredDistrLoss,
)
from skpro.metrics.survival import SPLL, ConcordanceHarrell



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Machine type converters for scitypes.

Exports
-------
convert_to(obj, to_type: str, as_scitype: str, store=None)
    converts object "obj" to type "to_type", considered as "as_scitype"

convert(obj, from_type: str, to_type: str, as_scitype: str, store=None)
    same as convert_to, without automatic identification of "from_type"

mtype(obj, as_scitype: str)
    returns "from_type" of obj, considered as "as_scitype"
---

Function signature of convert

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
from_type : str - the type to convert "obj" to, a valid mtype string
to_type : str - the type to convert "obj" to, a valid mtype string
as_scitype : str - name of scitype the object "obj" is considered as
store : reference of storage for lossy conversions, default=None (no store)

Returns
-------
converted_obj : to_type - object obj converted to to_type

---

Function signature of convert_to

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
to_type : str - the type to convert "obj" to, a valid mtype string
as_scitype : str - name of scitype the object "obj" is considered as
store : reference of storage for lossy conversions, default=None (no store)

Returns
-------
converted_obj : to_type - object obj converted to to_type

---

Function signature of mtype

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
as_scitype : str - name of scitype the object "obj" is considered as

Returns
-------
str - the type to convert "obj" to, a valid mtype string
    or None, if obj is None
"""

__author__ = ["fkiraly"]

__all__ = [
    "convert",
    "convert_to",
]

from copy import deepcopy
from functools import lru_cache

import numpy as np
import pandas as pd

from skpro.datatypes._base import BaseConverter
from skpro.datatypes._check import mtype as infer_mtype
from skpro.datatypes._proba import convert_dict_Proba
from skpro.datatypes._registry import mtype_to_scitype
from skpro.datatypes._table import convert_dict_Table


def get_convert_dict(soft_deps="present"):
    """Retrieve convert_dict, caches the first time it is requested.

    This is to avoid repeated, time consuming crawling in generate_check_dict,
    which would otherwise be called every time check_dict is requested.

    Parameters
    ----------
    soft_deps : str, optional - one of "present", "all"
        "present" - only conversions with soft dependencies present are included
        "all" - all conversions are included
    """
    if soft_deps not in ["present", "all"]:
        raise ValueError(
            "Error in get_check_dict, soft_deps argument must be 'present' or 'all', "
            f"found {soft_deps}"
        )
    convert_dict = generate_convert_dict(soft_deps=soft_deps)
    return convert_dict.copy()


@lru_cache(maxsize=1)
def generate_convert_dict(soft_deps="present"):
    """Generate convert_dict using lookup."""
    from skbase.utils.dependencies import _check_estimator_deps

    from skpro.utils.retrieval import _all_classes

    classes = _all_classes("skpro.datatypes")
    classes = [x[1] for x in classes]
    classes = [x for x in classes if issubclass(x, BaseConverter)]
    classes = [x for x in classes if not x.__name__.startswith("Base")]

    # subset only to data types with soft dependencies present
    if soft_deps == "present":
        classes = [x for x in classes if _check_estimator_deps(x, severity="none")]

    convert_dict = dict()
    for cls in classes:
        if not cls.get_class_tag("multiple_conversions", False):
            k = cls()
            key = k._get_key()
            convert_dict[key] = k
        else:
            for cls_to_cls in cls.get_conversions():
                k = cls(*cls_to_cls)

                # check dependencies for both classes
                # only add conversions if dependencies are satisfied for to and from
                cls_from, cls_to = k._get_cls_from_to()

                # do not add conversion if dependencies are not satisfied
                if cls_from is None or cls_to is None:
                    continue
                filter_sd = soft_deps in ["present"]
                if filter_sd and not _check_estimator_deps(cls_from, severity="none"):
                    continue
                if filter_sd and not _check_estimator_deps(cls_to, severity="none"):
                    continue

                key = k._get_key()
                convert_dict[key] = k

    # temporary while refactoring
    convert_dict.update(convert_dict_Proba)
    convert_dict.update(convert_dict_Table)

    return convert_dict


def convert(
    obj,
    from_type: str,
    to_type: str,
    as_scitype: str = None,
    store=None,
    store_behaviour: str = None,
    return_to_mtype: bool = False,
):
    """Convert objects between different machine representations, subject to scitype.

    Parameters
    ----------
    obj : object to convert - any type, should comply with mtype spec for as_scitype
    from_type : str - the type to convert "obj" to, a valid mtype string
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    to_type : str - the mtype to convert "obj" to, a valid mtype string
        or list of str, this specifies admissible types for conversion to;
        if list, will convert to first mtype of the same scitype as from_mtype
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    as_scitype : str, optional - name of scitype the object "obj" is considered as
        default = inferred from from_type
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
    store : optional, reference of storage for lossy conversions, default=None (no ref)
        is updated by side effect if not None and store_behaviour="reset" or "update"
    store_behaviour : str, optional, one of None (default), "reset", "freeze", "update"
        "reset" - store is emptied and then updated from conversion
        "freeze" - store is read-only, may be read/used by conversion but not changed
        "update" - store is updated from conversion and retains previous contents
        None - automatic: "update" if store is empty and not None; "freeze", otherwise
    return_to_mtype: bool, optional (default=False)
        if True, also returns the str of the mtype converted to

    Returns
    -------
    converted_obj : to_type - object obj converted to to_type
                    if obj was None, returns None
    to_type : str, only returned if ``return_to_mtype=True``
        mtype of ``converted_obj`` - useful of ``to_type`` was a list

    Raises
    ------
    KeyError if conversion is not implemented
    TypeError or ValueError if inputs do not match specification
    """
    if obj is None:
        return None

    # if to_type is a list, we do the following:
    # if on the list, then don't do a conversion (convert to from_type)
    # if not on the list, we find and convert to first mtype that has same scitype
    to_type = _get_first_mtype_of_same_scitype(
        from_mtype=from_type, to_mtypes=to_type, varname="to_type"
    )

    # input type checks
    if not (
        isinstance(to_type, list) and all(isinstance(item, str) for item in to_type)
    ) and not isinstance(to_type, str):
        raise TypeError("to_type must be a str or list of str")
    if not isinstance(from_type, str):
        raise TypeError("from_type must be a str")
    if as_scitype is None:
        as_scitype = mtype_to_scitype(to_type)
    elif not isinstance(as_scitype, str):
        raise TypeError("as_scitype must be str or None")
    if store is not None and not isinstance(store, dict):
        raise TypeError("store must be a dict or None")
    if store_behaviour not in [None, "reset", "freeze", "update"]:
        raise ValueError(
            'store_behaviour must be one of "reset", "freeze", "update", or None'
        )
    if store_behaviour is None and store == {}:
        store_behaviour = "update"
    if store_behaviour is None and store != {}:
        store_behaviour = "freeze"

    key = (from_type, to_type, as_scitype)

    convert_dict = get_convert_dict()
    if key not in convert_dict.keys():
        raise NotImplementedError(
            "no conversion defined from type " + str(from_type) + " to " + str(to_type)
        )

    if store_behaviour == "freeze":
        store = deepcopy(store)
    elif store_behaviour == "reset":
        # note: this is a side effect on store
        store.clear()
    elif store_behaviour == "update":
        # store is passed to convert_obj by reference, unchanged
        # this "elif" is here for clarity, to cover all three values
        pass
    else:
        raise RuntimeError(
            "bug: unreachable condition error, store_behaviour has unexpected value"
        )

    converted_obj = convert_dict[key](obj, store=store)

    if return_to_mtype:
        return converted_obj, to_type
    else:
        return converted_obj


# conversion based on queryable type to specified target
def convert_to(
    obj,
    to_type: str,
    as_scitype: str = None,
    store=None,
    store_behaviour: str = None,
    return_to_mtype: bool = False,
):
    """Convert object to a different machine representation, subject to scitype.

    Parameters
    ----------
    obj : object to convert - any type, should comply with mtype spec for as_scitype
    to_type : str - the mtype to convert "obj" to, a valid mtype string
            or list of str, this specifies admissible types for conversion to;
            if list, will convert to first mtype of the same scitype as obj
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    as_scitype : str, optional - name of scitype the object "obj" is considered as
        pre-specifying the scitype reduces the number of checks done in type inference
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
        default = inferred from mtype of obj, which is in turn inferred internally
    store : reference of storage for lossy conversions, default=None (no store)
        is updated by side effect if not None and store_behaviour="reset" or "update"
    store_behaviour : str, optional, one of None (default), "reset", "freeze", "update"
        "reset" - store is emptied and then updated from conversion
        "freeze" - store is read-only, may be read/used by conversion but not changed
        "update" - store is updated from conversion and retains previous contents
        None - automatic: "update" if store is empty and not None; "freeze", otherwise
    return_to_mtype: bool, optional (default=False)
        if True, also returns the str of the mtype converted to
    Returns
    -------
    converted_obj : to_type - object obj, or obj converted to target mtype as follows:
        case 1: mtype of obj is equal to to_type, or a list element of to_type
            no conversion happens, converted_obj = obj
        case 2: to_type is a str, and not equal to mtype of obj
            converted_obj is obj converted to to_type
        case 3: to_type is list of str, and mtype of obj is not in that list
            converted_obj is converted to the first mtype in to_type
                that is of same scitype as obj
        case 4: if obj was None, converted_obj is also None
    to_type : str, only returned if ``return_to_mtype=True``
        mtype of ``converted_obj`` - useful of ``to_type`` was a list

    Raises
    ------
    TypeError if machine type of input "obj" is not recognized
    TypeError if to_type contains no mtype compatible with mtype of obj
    KeyError if conversion that would be conducted is not implemented
    TypeError or ValueError if inputs do not match specification
    """
    if obj is None:
        return None

    # input checks on to_type, as_scitype; coerce to_type, as_scitype to lists
    to_type = _check_str_or_list_of_str(to_type, obj_name="to_type")

    # sub-set a preliminary set of as_scitype from to_type, as_scitype
    if as_scitype is not None:
        # if not None, subset to types compatible between to_type and as_scitype
        as_scitype = _check_str_or_list_of_str(as_scitype, obj_name="as_scitype")
        potential_scitypes = mtype_to_scitype(to_type)
        as_scitype = list(set(potential_scitypes).intersection(as_scitype))
    else:
        # if None, infer from to_type
        as_scitype = mtype_to_scitype(to_type)

    # now further narrow down as_scitype by inference from the obj
    from_type = infer_mtype(obj=obj, as_scitype=as_scitype)
    as_scitype = mtype_to_scitype(from_type)

    converted_obj = convert(
        obj=obj,
        from_type=from_type,
        to_type=to_type,
        as_scitype=as_scitype,
        store=store,
        store_behaviour=store_behaviour,
        return_to_mtype=return_to_mtype,
    )

    return converted_obj


def _get_first_mtype_of_same_scitype(from_mtype, to_mtypes, varname="to_mtypes"):
    """Return first mtype in list mtypes that has same scitype as from_mtype.

    Parameters
    ----------
    from_mtype : str - mtype of object to convert from
    to_mtypes : list of str - mtypes to convert to
    varname : str - name of variable to_mtypes, for error message

    Returns
    -------
    to_type : str - first mtype in to_mtypes that has same scitype as from_mtype
    """
    to_mtypes = _check_str_or_list_of_str(to_mtypes, obj_name=varname)
    if not isinstance(to_mtypes, list):
        raise TypeError(f"{varname} must be a str or a list of str")
    # no conversion of from_type is in the list
    if from_mtype in to_mtypes:
        return from_mtype
    # otherwise convert to first element of same scitype
    scitype = mtype_to_scitype(from_mtype)
    same_scitype_mtypes = [
        mtype for mtype in to_mtypes if mtype_to_scitype(mtype) == scitype
    ]
    if len(same_scitype_mtypes) == 0:
        raise TypeError(
            f"{varname} contains no mtype compatible with the scitype of obj, "
            f"which is {scitype}. Value of {varname} is: {to_mtypes}"
        )
    to_type = same_scitype_mtypes[0]
    return to_type


def _conversions_defined(scitype: str, soft_deps: str = "present"):
    """Return an indicator matrix which conversions are defined for scitype.

    Parameters
    ----------
    scitype: str - name of scitype for which conversions are queried
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
    soft_deps : str, optional - one of "present", "all"
        "present" - only conversions with soft dependencies present are included
        "all" - all conversions are included

    Returns
    -------
    conv_df: pd.DataFrame, columns and index is list of mtypes for scitype
            entry of row i, col j is 1 if conversion from i to j is defined,
                                     0 if conversion from i to j is not defined
    """
    convert_dict = get_convert_dict(soft_deps=soft_deps)
    pairs = [(x[0], x[1]) for x in list(convert_dict.keys()) if x[2] == scitype]
    cols0 = {x[0] for x in list(convert_dict.keys()) if x[2] == scitype}
    cols1 = {x[1] for x in list(convert_dict.keys()) if x[2] == scitype}
    cols = sorted(list(cols0.union(cols1)))

    mat = np.zeros((len(cols), len(cols)), dtype=int)
    nkeys = len(cols)
    for i in range(nkeys):
        for j in range(nkeys):
            if (cols[i], cols[j]) in pairs:
                mat[i, j] = 1

    conv_df = pd.DataFrame(mat, index=cols, columns=cols)

    return conv_df


def _check_str_or_list_of_str(obj, obj_name="obj"):
    """Check whether obj is str or list of str; coerces to list of str."""
    if isinstance(obj, list):
        if not np.all([isinstance(x, str) for x in obj]):
            raise TypeError(f"{obj} must be a str or list of str")
        else:
            return obj
    elif not isinstance(obj, str):
        raise TypeError(f"{obj} must be a str or list of str")
    else:
        return [obj]



"""Estimator checker for extension."""

__author__ = ["fkiraly"]
__all__ = ["check_estimator"]

from skbase.utils.dependencies import _check_soft_dependencies


def check_estimator(
    estimator,
    raise_exceptions=False,
    tests_to_run=None,
    fixtures_to_run=None,
    verbose=True,
    tests_to_exclude=None,
    fixtures_to_exclude=None,
):
    """Run all tests on one single estimator.

    Tests that are run on estimator:
        all tests in test_all_estimators
        all interface compatibility tests from the module of estimator's scitype
            for example, test_all_regressors if estimator is a regressor

    Parameters
    ----------
    estimator : estimator class or estimator instance
    raise_exceptions : bool, optional, default=False
        whether to return exceptions/failures in the results dict, or raise them

        * if False: returns exceptions in returned `results` dict
        * if True: raises exceptions as they occur

    tests_to_run : str or list of str, optional. Default = run all tests.
        Names (test/function name string) of tests to run.
        sub-sets tests that are run to the tests given here.
    fixtures_to_run : str or list of str, optional. Default = run all tests.
        pytest test-fixture combination codes, which test-fixture combinations to run.
        sub-sets tests and fixtures to run to the list given here.
        If both tests_to_run and fixtures_to_run are provided, runs the *union*,
        i.e., all test-fixture combinations for tests in tests_to_run,
            plus all test-fixture combinations in fixtures_to_run.
    verbose : str, optional, default=True.
        whether to print out informative summary of tests run.
    tests_to_exclude : str or list of str, names of tests to exclude. default = None
        removes tests that should not be run, after subsetting via tests_to_run.
    fixtures_to_exclude : str or list of str, fixtures to exclude. default = None
        removes test-fixture combinations that should not be run.
        This is done after subsetting via fixtures_to_run.

    Returns
    -------
    results : dict of results of the tests in self
        keys are test/fixture strings, identical as in pytest, e.g., test[fixture]
        entries are the string "PASSED" if the test passed,
        or the exception raised if the test did not pass
        returned only if all tests pass, or raise_exceptions=False

    Raises
    ------
    if raise_exceptions=True,
    raises any exception produced by the tests directly

    Examples
    --------
    >>> from skpro.regression.residual import ResidualDouble
    >>> from skpro.utils import check_estimator

    Running all tests for ResidualDouble class,
    this uses all instances from get_test_params and compatible scenarios

    >>> results = check_estimator(ResidualDouble)
    All tests PASSED!

    Running all tests for a specific ResidualDouble
    this uses the instance that is passed and compatible scenarios

    >>> from sklearn.linear_model import LinearRegression
    >>> results = check_estimator(ResidualDouble(LinearRegression()))
    All tests PASSED!

    Running specific test (all fixtures) for ResidualDouble

    >>> results = check_estimator(ResidualDouble, tests_to_run="test_clone")
    All tests PASSED!

    {'test_clone[ResidualDouble-0]': 'PASSED',
    'test_clone[ResidualDouble-1]': 'PASSED'}

    Running one specific test-fixture-combination for ResidualDouble

    >>> check_estimator(
    ...    ResidualDouble, fixtures_to_run="test_clone[ResidualDouble-1]"
    ... )
    All tests PASSED!
    {'test_clone[ResidualDouble-1]': 'PASSED'}
    """
    msg = (
        "check_estimator is a testing utility for developers, and "
        "requires pytest to be present "
        "in the python environment, but pytest was not found. "
        "pytest is a developer dependency and not included in the base "
        "sktime installation. Please run: `pip install pytest` to "
        "install the pytest package. "
        "To install sktime with all developer dependencies, run:"
        " `pip install sktime[dev]`"
    )
    _check_soft_dependencies("pytest", msg=msg)

    from skpro.tests.test_class_register import get_test_classes_for_obj

    test_clss_for_est = get_test_classes_for_obj(estimator)

    results = {}

    for test_cls in test_clss_for_est:
        test_cls_results = test_cls().run_tests(
            obj=estimator,
            raise_exceptions=raise_exceptions,
            tests_to_run=tests_to_run,
            fixtures_to_run=fixtures_to_run,
            tests_to_exclude=tests_to_exclude,
            fixtures_to_exclude=fixtures_to_exclude,
        )
        results.update(test_cls_results)

    failed_tests = [key for key in results.keys() if results[key] != "PASSED"]
    if len(failed_tests) > 0:
        msg = failed_tests
        msg = ["FAILED: " + x for x in msg]
        msg = "\n".join(msg)
    else:
        msg = "All tests PASSED!"

    if verbose:
        # printing is an intended feature, for console usage and interactive debugging
        print(msg)  # noqa T001

    return results



"""Delegator mixin that delegates all methods to wrapped distribution.

Useful for building estimators where all but one or a few methods are delegated. For
that purpose, inherit from this estimator and then override only the methods that
are not delegated.
"""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]
__all__ = ["_DelegatedDistribution"]

from copy import deepcopy

from skpro.distributions.base._base import BaseDistribution


class _DelegatedDistribution(BaseDistribution):
    """Delegator mixin that delegates all methods to wrapped estimator.

    Delegates inner methods to a wrapped estimator.
        Wrapped estimator is value of attribute with name self._delegate_name.
        By default, this is "estimator_", i.e., delegates to self.estimator_
        To override delegation, override _delegate_name attribute in child class.

    Delegates the following methods:
        _iloc, pdf, log_pdf, cdf, ppf, energy, mean, var, pdfnorm, sample

    Does NOT delegate get_params, set_params.
        get_params, set_params will hence use one additional nesting level by default.

    Does NOT delegate or copy tags, this should be done in a child class if required.
    """

    # attribute for _DelegatedDistribution, which then delegates
    #     all non-overridden methods are same as of getattr(self, _delegate_name)
    #     see further details in _DelegatedDistribution docstring
    _delegate_name = "delegate_"

    def _get_delegate(self):
        return getattr(self, self._delegate_name)

    def _iloc(self, rowidx=None, colidx=None):
        cls = self.__class__

        delegate = self._get_delegate()
        delegate_subset = delegate.iloc[rowidx, colidx]
        delegate_subset_params = deepcopy(delegate_subset.get_params())

        return cls(**delegate_subset_params)

    def pdf(self, x):
        r"""Probability density function.

        Let :math:`X` be a random variables with the distribution of `self`,
        taking values in `(N, n)` `DataFrame`-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input `x` representing :math:`x`,
        is a `DataFrame` with same columns and indices as `self`,
        and entries :math:`p_{X_{ij}}(x_{ij})`.

        Parameters
        ----------
        x : `pandas.DataFrame` or 2D np.ndarray
            representing :math:`x`, as above

        Returns
        -------
        `DataFrame` with same columns and index as `self`
            containing :math:`p_{X_{ij}}(x_{ij})`, as above
        """
        delegate = self._get_delegate()
        return delegate.pdf(x)

    def log_pdf(self, x):
        r"""Logarithmic probability density function.

        Numerically more stable than calling pdf and then taking logartihms.

        Let :math:`X` be a random variables with the distribution of `self`,
        taking values in `(N, n)` `DataFrame`-s
        Let :math:`x\in \mathbb{R}^{N\times n}`.
        By :math:`p_{X_{ij}}`, denote the marginal pdf of :math:`X` at the
        :math:`(i,j)`-th entry.

        The output of this method, for input `x` representing :math:`x`,
        is a `DataFrame` with same columns and indices as `self`,
        and entries :math:`\log p_{X_{ij}}(x_{ij})`.

        If `self` has a mixed or discrete distribution, this returns
        the weighted continuous part of `self`'s distribution instead of the pdf,
        i.e., the marginal pdf integrate to the weight of the continuous part.

        Parameters
        ----------
        x : `pandas.DataFrame` or 2D np.ndarray
            representing :math:`x`, as above

        Returns
        -------
        `DataFrame` with same columns and index as `self`
            containing :math:`\log p_{X_{ij}}(x_{ij})`, as above
        """
        delegate = self._get_delegate()
        return delegate.log_pdf(x)

    def cdf(self, x):
        """Cumulative distribution function."""
        delegate = self._get_delegate()
        return delegate.cdf(x)

    def ppf(self, p):
        """Quantile function = percent point function = inverse cdf."""
        delegate = self._get_delegate()
        return delegate.ppf(p)

    def energy(self, x=None):
        r"""Energy of self, w.r.t. self or a constant frame x.

        Let :math:`X, Y` be i.i.d. random variables with the distribution of `self`.

        If `x` is `None`, returns :math:`\mathbb{E}[|X-Y|]` (for each row),
        "self-energy" (of the row marginal distribution).
        If `x` is passed, returns :math:`\mathbb{E}[|X-x|]` (for each row),
        "energy wrt x" (of the row marginal distribution).

        Parameters
        ----------
        x : None or pd.DataFrame, optional, default=None
            if pd.DataFrame, must have same rows and columns as `self`

        Returns
        -------
        pd.DataFrame with same rows as `self`, single column `"energy"`
        each row contains one float, self-energy/energy as described above.
        """
        delegate = self._get_delegate()
        return delegate.energy(x=x)

    def mean(self):
        r"""Return expected value of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns the expectation :math:`\mathbb{E}[X]`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        expected value of distribution (entry-wise)
        """
        delegate = self._get_delegate()
        return delegate.mean()

    def var(self):
        r"""Return element/entry-wise variance of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns :math:`\mathbb{V}[X] = \mathbb{E}\left(X - \mathbb{E}[X]\right)^2`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        variance of distribution (entry-wise)
        """
        delegate = self._get_delegate()
        return delegate.var()

    def pdfnorm(self, a=2):
        r"""a-norm of pdf, defaults to 2-norm.

        computes a-norm of the entry marginal pdf, i.e.,
        :math:`\mathbb{E}[p_X(X)^{a-1}] = \int p(x)^a dx`,
        where :math:`X` is a random variable distributed according to the entry marginal
        of `self`, and :math:`p_X` is its pdf

        Parameters
        ----------
        a: int or float, optional, default=2

        Returns
        -------
        pd.DataFrame with same rows and columns as `self`
        each entry is :math:`\mathbb{E}[p_X(X)^{a-1}] = \int p(x)^a dx`, see above
        """
        delegate = self._get_delegate()
        return delegate.pdfnorm(a=a)

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        delegate = self._get_delegate()
        return delegate.sample(n_samples=n_samples)



"""Extension template for time-to-event predictors aka survival predictors.

Purpose of this implementation template:
    quick implementation of new estimators following the template
    NOT a concrete class to import! This is NOT a base class or concrete class!
    This is to be used as a "fill-in" coding template.

How to use this implementation template to implement a new distribution:
- make a copy of the template in a suitable location, give it a descriptive name.
- work through all the "todo" comments below
- fill in code for mandatory methods, and optionally for optional methods
- do not write to reserved attributes: is_fitted, _is_fitted, _X_metadata, _y_metadata,
  _tags, _tags_dynamic, _config, _config_dynamic
- you can add more private methods, but do not override BaseEstimator's private methods
    an easy way to be safe is to prefix your methods with "_custom"
- change docstrings for functions and the file
- ensure interface compatibility by skpro.utils.estimator_checks.check_estimator
- once complete: use as a local library, or contribute to skpro via PR
- more details:
  https://www.sktime.net/en/stable/developer_guide/add_estimators.html

Mandatory methods to implement:
    fitting         - _fit(self, X, y, C=None)

The method must handle the case C=None, as well as the case where C is a pd.DataFrame.

At least one of the following probabilistic prediction methods:
    predicting quantiles         - _predict_quantiles(self, X, alpha=None)
    OR predicting intervals      - _predict_interval(self, X, coverage=None)
    OR predicting distribution   - _predict_proba(self, X)

Optional methods to implement:
    predicting variance          - _predict_var(self, X, cov=False)
    updating                     - _update(self, X, y)

Testing - required for test framework and check_estimator usage:
    get default parameters for test instance(s) - get_test_params()
"""
# todo: write an informative docstring for the file or module, remove the above
# todo: add an appropriate copyright notice for your estimator
#       estimators contributed to skpro should have the copyright notice at the top
#       estimators of your own do not need to have permissive or BSD-3 copyright

# todo: uncomment the following line, enter authors' GitHub IDs
# __author__ = [authorGitHubID, anotherAuthorGitHubID]

from skpro.survival.base import BaseSurvReg

# todo: add any necessary imports here

# todo: for imports of skpro soft dependencies:
# make sure to fill in the "python_dependencies" tag with the package import name
# import soft dependencies only inside methods of the class, not at the top of the file


# todo: change class name and write docstring
class ClassName(BaseSurvReg):
    """Custom probabilistic survival regressor. todo: write docstring.

    todo: describe your custom regressor here

    Parameters
    ----------
    parama : int
        descriptive explanation of parama
    paramb : string, optional (default='default')
        descriptive explanation of paramb
    paramc : boolean, optional (default= whether paramb is not the default)
        descriptive explanation of paramc
    and so on
    est : skpro.estimator, BaseEstimator descendant
        descriptive explanation of est
    est2: another estimator
        descriptive explanation of est2
    and so on
    """

    # todo: fill out estimator tags here
    #  tags are inherited from parent class if they are not set
    # tags inherited from base are "safe defaults" which can usually be left as-is
    _tags = {"capability:survival": True}

    # todo: fill init
    # params should be written to self and never changed
    # super call must not be removed, change class name
    # parameter checks can go after super call
    def __init__(self, paramname, paramname2="paramname2default"):
        # estimators should precede parameters
        #  if estimators have default values, set None and initialize below

        # todo: write any hyper-parameters and components to self
        self.paramname = paramname
        self.paramname2 = paramname2

        # leave this as is
        super().__init__()

        # todo: optional, parameter checking logic (if applicable) should happen here
        # if writes derived values to self, should *not* overwrite self.parama etc
        # instead, write to self._parama, self._newparam (starting with _)

        # todo: default estimators should have None arg defaults
        #  and be initialized here
        #  do this only with default estimators, not with parameters
        # if est2 is None:
        #     self.estimator = MyDefaultEstimator()

        # todo: if tags of estimator depend on component tags, set these here
        #  only needed if estimator is a composite
        #  tags set in the constructor apply to the object and override the class
        #
        # example 1: conditional setting of a tag
        # if est.foo == 42:
        #   self.set_tags(handles-missing-data=True)
        # example 2: cloning tags from component
        #   self.clone_tags(est2, ["enforce_index_type", "handles-missing-data"])

    # todo: implement this, mandatory
    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Changes state to "fitted" = sets is_fitted flag to True

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
        """
        # insert logic for estimator here
        # fitted parameters should be written to parameters ending in underscore

        # the estimator should be able to cover
        # the case where C is a pd.DataFrame,
        # and the case where C is None (interpreted as "no censoring")

        # self must be returned at the end
        return self

    # todo: implement this, mandatory
    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        # implement logic for prediction here
        # this can read out parameters fitted in fit, or hyperparameters from init
        # no attributes should be written to self

        y_pred = "placeholder"
        # returned object should be pd.DataFrame
        # same length as X, same columns as y in fit
        return y_pred

    # todo: implement at least one of the probabilistic prediction methods
    # _predict_proba, _predict_interval, _predict_quantiles
    # if one is implemented, the other two are filled in by default
    # implementation of _predict_proba is preferred, if possible
    #
    # CAVEAT: if not implemented, _predict_proba assumes normal distribution
    # this can be inconsistent with _predict_interval or _predict_quantiles
    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        # if implementing _predict_proba (otherwise delete this method)
        # todo: adapt the following by filling in logic to produce prediction values

        # boilerplate code to create correct output index
        index = X.index
        y_cols = self._y_cols  # columns from y in fit, not automatically stored
        columns = y_cols

        # values = logic to produce prediction values
        # replace this import by the distribution you are using
        # the distribution type can be conditional, e.g., data or parameter dependent
        from skpro.distributions import SomeDistribution

        values = None  # fill in values
        y_pred = SomeDistribution(values, index=index, columns=columns)

        return y_pred

    # todo: implement at least one of the probabilistic prediction methods, see above
    # delete the methods that are not implemented and filled by default
    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        # if implementing _predict_interval (otherwise delete this method)
        # todo: adapt the following by filling in logic to produce prediction values

        # boilerplate code to create correct pandas output index
        # only if using pandas, for other mtypes, use appropriate data structure
        import pandas as pd

        index = X.index
        y_cols = self._y_cols  # columns from y in fit, not automatically stored
        columns = pd.MultiIndex.from_product(
            [y_cols, coverage, ["lower", "upper"]],
        )

        # values = logic to produce prediction values
        values = None  # fill in values
        pred_int = pd.DataFrame(values, index=index, columns=columns)

        return pred_int

    # todo: implement at least one of the probabilistic prediction methods, see above
    # delete the methods that are not implemented and filled by default
    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        # if implementing _predict_quantiles (otherwise delete this method)
        # todo: adapt the following by filling in logic to produce prediction values

        # boilerplate code to create correct pandas output index
        # only if using pandas, for other mtypes, use appropriate data structure
        import pandas as pd

        index = X.index
        y_cols = self._y_cols  # columns from y in fit, not automatically stored
        columns = pd.MultiIndex.from_product(
            [y_cols, alpha],
        )

        # values = logic to produce prediction values
        values = None  # fill in values
        quantiles = pd.DataFrame(values, index=index, columns=columns)

        return quantiles

    # todo: return default parameters, so that a test instance can be created
    #   required for automated unit and integration testing of estimator
    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """

        # todo: set the testing parameters for the estimators
        # Testing parameters can be dictionary or list of dictionaries
        #
        # this can, if required, use:
        #   class properties (e.g., inherited); parent class test case
        #   imported objects such as estimators from skpro or sklearn
        # important: all such imports should be *inside get_test_params*, not at the top
        #            since imports are used only at testing time
        #
        # The parameter_set argument is not used for most automated, module level tests.
        #   It can be used in custom, estimator specific tests, for "special" settings.
        # A parameter dictionary must be returned *for all values* of parameter_set,
        #   i.e., "parameter_set not available" errors should never be raised.
        #
        # A good parameter set should primarily satisfy two criteria,
        #   1. Chosen set of parameters should have a low testing time,
        #      ideally in the magnitude of few seconds for the entire test suite.
        #       This is vital for the cases where default values result in
        #       "big" models which not only increases test time but also
        #       run into the risk of test workers crashing.
        #   2. There should be a minimum two such parameter sets with different
        #      sets of values to ensure a wide range of code coverage is provided.
        #
        # example 1: specify params as dictionary
        # any number of params can be specified
        # params = {"est": value0, "parama": value1, "paramb": value2}
        #
        # example 2: specify params as list of dictionary
        # note: Only first dictionary will be used by create_test_instance
        # params = [{"est": value1, "parama": value2},
        #           {"est": value3, "parama": value4}]
        #
        # example 3: parameter set depending on param_set value
        #   note: only needed if a separate parameter set is needed in tests
        # if parameter_set == "special_param_set":
        #     params = {"est": value1, "parama": value2}
        #     return params
        #
        # # "default" params
        # params = {"est": value3, "parama": value4}
        # return params



# LEGACY MODULE - TODO: remove or refactor

import abc
import hashlib
import json

from sklearn.base import clone


class Model:
    """Model base class

    Parameters
    ----------
    instance
    tuning
    group
    database
    """

    def __init__(self, instance, tuning=None, group=None, name=None, database=None):
        self.instance = clone(instance)
        self.group = group
        if database is None:
            database = {}
        self.database = database
        if isinstance(tuning, dict) and len(tuning) == 0:
            tuning = None
        self.tuning = tuning
        self.name = name

    def __repr__(self):
        return "Model(" + repr(self.instance) + ")"

    def __str__(self):
        if isinstance(self.name, str):
            return "Model: " + self.name
        return "Model(" + str(self.instance) + ")"

    def __getitem__(self, item):
        return self.database[item]

    def __setitem__(self, key, value):
        self.database[key] = value

    def clone(self):
        return Model(self.instance, self.tuning, self.group)

    def identifier(self, with_tuning=False):
        s = repr(self.instance)
        if with_tuning:
            s += json.dumps(self.tuning, sort_keys=True)

        return hashlib.md5(s.encode("utf-8")).hexdigest()


class Controller(metaclass=abc.ABCMeta):
    """Controller base class

    Notes
    -----
    Must implement an identifier and run function
    """

    @abc.abstractmethod
    def identifier(self):
        pass

    @abc.abstractmethod
    def run(self, model):
        pass


class View(metaclass=abc.ABCMeta):
    """View base class

    Notes
    -----
    Must implement the parse method
    """

    def __repr__(self):
        return self.__class__.__name__

    @abc.abstractmethod
    def parse(self, data):
        pass



"""Non-suite tests for probability distribution objects."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd
import pytest
from skbase.utils.dependencies import _check_soft_dependencies

from skpro.tests.test_switch import run_test_module_changed


@pytest.mark.skipif(
    not run_test_module_changed("skpro.distributions"),
    reason="run only if skpro.distributions has been changed",
)
def test_proba_example():
    """Test one subsetting case for BaseDistribution."""
    from skpro.distributions.normal import Normal

    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)

    assert n.shape == (3, 2)

    one_row = n.loc[[1]]
    assert isinstance(one_row, Normal)
    assert one_row.shape == (1, 2)


@pytest.mark.parametrize("subsetter", ["loc", "iloc"])
def test_proba_subsetters_loc_iloc(subsetter):
    """Test one subsetting case for BaseDistribution."""
    from skpro.distributions.normal import Normal

    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)

    assert n.shape == (3, 2)

    # should result in 2D array distribution (1, 1)
    nss = getattr(n, subsetter)[1, [1]]
    assert isinstance(nss, Normal)
    assert nss.shape == (1, 1)
    assert nss.mu.shape == (1, 1)

    nss = getattr(n, subsetter)[[1], 1]
    assert isinstance(nss, Normal)
    assert nss.shape == (1, 1)
    assert nss.mu.shape == (1, 1)

    # should result in scalar distribution
    nss = getattr(n, subsetter)[1, 1]
    assert isinstance(nss, Normal)
    assert nss.shape == ()

    nss = getattr(n, subsetter)[1, 1]
    assert isinstance(nss, Normal)
    assert nss.shape == ()


def test_proba_subsetters_at_iat():
    """Test one subsetting case for BaseDistribution."""
    from skpro.distributions.normal import Normal

    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)

    # should result in scalar distribution
    nss = n.iat[1, 1]
    assert isinstance(nss, Normal)
    assert nss.shape == ()
    assert nss == n.iloc[1, 1]

    nss = n.at[1, 1]
    assert isinstance(nss, Normal)
    assert nss.shape == ()
    assert nss == n.loc[1, 1]


def test_proba_index_coercion():
    """Test index coercion for BaseDistribution."""
    from skpro.distributions.normal import Normal

    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, columns=["foo", "bar"])

    assert n.shape == (3, 2)
    assert isinstance(n.index, pd.Index)
    assert isinstance(n.columns, pd.Index)
    assert n.index.equals(pd.RangeIndex(3))
    assert n.columns.equals(pd.Index(["foo", "bar"]))

    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, index=["2", 1, 0])

    assert n.shape == (3, 2)
    assert isinstance(n.index, pd.Index)
    assert isinstance(n.columns, pd.Index)
    assert n.index.equals(pd.Index(["2", 1, 0]))
    assert n.columns.equals(pd.RangeIndex(2))

    # this should coerce to a 2D array of shape (1, 3)
    n = Normal(0, 1, columns=[1, 2, 3])

    assert n.shape == (1, 3)
    assert isinstance(n.index, pd.Index)
    assert isinstance(n.columns, pd.Index)
    assert n.index.equals(pd.RangeIndex(1))
    assert n.columns.equals(pd.Index([1, 2, 3]))


@pytest.mark.skipif(
    not _check_soft_dependencies("matplotlib", severity="none"),
    reason="skip if matplotlib is not available",
)
@pytest.mark.parametrize("fun", ["pdf", "ppf", "cdf"])
def test_proba_plotting(fun):
    """Test that plotting functions do not crash and return ax as expected."""
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure

    from skpro.distributions.normal import Normal

    # default case, 2D distribution with n_columns>1
    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)
    fig, ax = n.plot(fun=fun)
    assert isinstance(fig, Figure)
    assert isinstance(ax, np.ndarray)
    assert ax.shape == n.shape
    assert all([isinstance(a, Axes) for a in ax.flatten()])
    assert all([a.get_figure() == fig for a in ax.flatten()])

    # 1D case requires special treatment of axes
    n = Normal(mu=[[1], [2], [3]], sigma=1)
    fig, ax = n.plot(fun=fun)
    assert isinstance(fig, Figure)
    assert isinstance(ax, type(ax))
    assert ax.shape == (n.shape[0],)
    assert all([isinstance(a, Axes) for a in ax.flatten()])
    assert all([a.get_figure() == fig for a in ax.flatten()])

    # scalar case
    n = Normal(mu=1, sigma=1)
    ax = n.plot(fun=fun)
    assert isinstance(ax, Axes)


def test_to_df_parametric():
    """Tests coercion to DataFrame via get_params_df and to_df."""
    from skpro.distributions.normal import Normal

    cols = ["foo", "bar"]

    # default case, 2D distribution with n_columns>1
    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, columns=cols)

    param_names = n.get_param_names()
    params_df = n.get_params_df()
    for k, v in params_df.items():
        assert k in param_names
        assert isinstance(v, pd.DataFrame)
        assert (v.index == n.index).all()
        assert (v.columns == n.columns).all()

    all_params_df = n.to_df()
    assert isinstance(all_params_df, pd.DataFrame)
    assert (all_params_df.index == n.index).all()
    assert isinstance(all_params_df.columns, pd.MultiIndex)

    level0_vals = all_params_df.columns.get_level_values(0).unique()
    level1_vals = all_params_df.columns.get_level_values(1).unique()

    assert (level0_vals == n.columns).all()
    for ix in level1_vals:
        assert ix in param_names
        assert ix not in ["index", "columns"]

    # scalar case
    n = Normal(mu=2, sigma=3)

    param_names = n.get_param_names()
    params_df = n.get_params_df()

    for k, v in params_df.items():
        assert k in param_names
        assert isinstance(v, pd.DataFrame)
        assert (v.index == pd.RangeIndex(1)).all()
        assert (v.columns == pd.RangeIndex(1)).all()

    all_params_df = n.to_df()
    assert isinstance(all_params_df, pd.DataFrame)
    assert (all_params_df.index == pd.RangeIndex(1)).all()
    assert not isinstance(all_params_df.columns, pd.MultiIndex)

    for ix in all_params_df.columns:
        assert ix in param_names
        assert ix not in ["index", "columns"]


def test_head_tail():
    """Test head and tail utility functions."""
    from skpro.distributions.normal import Normal

    cols = ["foo", "bar"]

    # default case, 2D distribution with n_columns>1
    n = Normal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, columns=cols)

    nh = n.head(2)
    assert isinstance(nh, Normal)
    assert nh.shape == (2, 2)
    assert (nh.columns == n.columns).all()
    assert (nh.index == n.index[:2]).all()

    nh2 = n.head()
    assert isinstance(nh2, Normal)
    assert nh2.shape == (3, 2)
    assert (nh2.columns == n.columns).all()
    assert (nh2.index == n.index).all()

    nt = n.tail(2)
    assert isinstance(nt, Normal)
    assert nt.shape == (2, 2)
    assert (nt.columns == n.columns).all()
    assert (nt.index == n.index[1:]).all()

    nt2 = n.tail()
    assert isinstance(nt2, Normal)
    assert nt2.shape == (3, 2)
    assert (nt2.columns == n.columns).all()
    assert (nt2.index == n.index).all()

    # scalar case
    n = Normal(mu=2, sigma=3)

    nh = n.head()
    assert nh.ndim == 0

    nt = n.tail(42)
    assert nt.ndim == 0



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Laplace probability distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Laplace(BaseDistribution):
    r"""Laplace distribution.

    This distribution is univariate, without correlation between dimensions
    for the array-valued case.

    The Laplace distribution is parametrized by mean :math:`\mu` and
    scale :math:`b`, such that the pdf is

    .. math:: f(x) = \frac{1}{2b} \exp\left(-\frac{|x - \mu|}{b}\right)

    The mean :math:`\mu` is represented by the parameter ``mu``,
    and the scale :math:`b` by the parameter ``scale``.

    It should be noted that this parametrization differs from the mean/standard
    deviation parametrization, which is also common in the literature.
    The standard deviation of this distribution is :math:`\sqrt{2} s`.

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        mean of the distribution
    scale : float or array of float (1D or 2D), must be positive
        scale parameter of the distribution, same as standard deviation / sqrt(2)
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions import Laplace

    >>> n = Laplace(mu=[[0, 1], [2, 3], [4, 5]], scale=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "energy", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, scale, index=None, columns=None):
        self.mu = mu
        self.scale = scale

        super().__init__(index=index, columns=columns)

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        energy_arr = self._bc_params["scale"]
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1) * 1.5
        return energy_arr

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        mu = self._bc_params["mu"]
        sc = self._bc_params["scale"]

        y_arr = np.abs((x - mu) / sc)
        c_arr = y_arr + np.exp(-y_arr)
        energy_arr = sc * c_arr
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._bc_params["mu"]

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        sc = self._bc_params["scale"]
        sd_arr = np.sqrt(2) * sc
        return sd_arr**2

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        mu = self._bc_params["mu"]
        sc = self._bc_params["scale"]
        pdf_arr = np.exp(-np.abs((x - mu) / sc))
        pdf_arr = pdf_arr / (2 * sc)
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        mu = self._bc_params["mu"]
        sc = self._bc_params["scale"]
        lpdf_arr = -np.abs((x - mu) / sc)
        lpdf_arr = lpdf_arr - np.log(2 * sc)
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        mu = self._bc_params["mu"]
        sc = self._bc_params["scale"]

        sgn_arr = np.sign(x - mu)
        exp_arr = np.exp(-np.abs((x - mu) / sc))
        cdf_arr = 0.5 + 0.5 * sgn_arr * (1 - exp_arr)
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        mu = self._bc_params["mu"]
        sc = self._bc_params["scale"]

        sgn_arr = np.sign(p - 0.5)
        icdf_arr = mu - sc * sgn_arr * np.log(1 - 2 * np.abs(p - 0.5))
        return icdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"mu": [[0, 1], [2, 3], [4, 5]], "scale": 1}
        params2 = {
            "mu": 0,
            "scale": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": -0.5, "scale": 3}

        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Log-Laplace probability distribution."""

__author__ = ["SaiRevanth25"]

import pandas as pd
from scipy.stats import loglaplace, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class LogLaplace(_ScipyAdapter):
    r"""Log-Laplace distribution.

    Most methods wrap ``scipy.stats.loglaplace``.

    This distribution is univariate, without correlation between dimensions
    for the array-valued case.

    The log-Laplace distribution is a continuous probability distribution obtained by
    taking the logarithm of the Laplace distribution, commonly used in finance and
    hydrology due to its heavy tails and asymmetry.

    The log-Laplace distribution is parametrized by the scale parameter
    :math:`\c`, such that the pdf is

    .. math:: f(x) = \frac{c}{2} x^{c-1}, \quad 0<x<1

    and

    .. math:: f(x) = \frac{c}{2} x^{-c-1}, \quad x >= 1

    The scale parameter :math:`c` is represented by the parameter ``c``.

    Parameters
    ----------
    scale : float or array of float (1D or 2D), must be positive
        scale parameter of the log-Laplace distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.loglaplace import LogLaplace

    >>> ll = LogLaplace(scale=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, scale, index=None, columns=None):
        self.scale = scale

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return loglaplace

    def _get_scipy_param(self):
        scale = self._bc_params["scale"]
        return [scale], {}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"scale": [[1, 2], [3, 4]]}
        params2 = {
            "scale": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"scale": 2}
        return [params1, params2, params3]



"""Test base class logic for scalar distributions.

Distributions behave polymorphically - if no index is passed,
all methods should accept and return scalars.

This is tested by using the Normal distribution as a test case,
which invokes the boilerplate for scalar distributions.
"""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd
import pytest

from skpro.distributions.normal import Normal
from skpro.tests.test_switch import run_test_module_changed


@pytest.mark.skipif(
    not run_test_module_changed("skpro.distributions"),
    reason="run only if skpro.distributions has been changed",
)
def test_scalar_distribution():
    """Test scalar distribution logic."""
    # test params
    mu = 1
    sigma = 2

    # instantiate distribution
    d = Normal(mu=mu, sigma=sigma)
    assert d.ndim == 0
    assert d.shape == ()
    assert d.index is None
    assert d.columns is None

    # test scalar input
    x = 0.5
    assert np.isscalar(d.mean())
    assert np.isscalar(d.var())
    assert np.isscalar(d.energy())
    assert np.isscalar(d.energy(x))
    assert np.isscalar(d.pdf(x))
    assert np.isscalar(d.log_pdf(x))
    assert np.isscalar(d.cdf(x))
    assert np.isscalar(d.ppf(x))
    assert np.isscalar(d.sample())

    spl_mult = d.sample(5)
    assert isinstance(spl_mult, pd.DataFrame)
    assert spl_mult.shape == (5, 1)
    assert spl_mult.index.equals(pd.RangeIndex(5))


@pytest.mark.skipif(
    not run_test_module_changed("skpro.distributions"),
    reason="run only if skpro.distributions has been changed",
)
def test_broadcast_ambiguous():
    """Test broadcasting in cases of ambiguous parameter dimensions."""
    mu = [1]
    sigma = 2
    # this should result in 2D array distribution
    # anything that is not scalar is broadcast to 2D
    d = Normal(mu=mu, sigma=sigma)
    assert d.ndim == 2
    assert d.shape == (1, 1)
    assert d.index.equals(pd.RangeIndex(1))
    assert d.columns.equals(pd.RangeIndex(1))

    def is_expected_2d(output, col=None):
        assert isinstance(output, pd.DataFrame)
        assert output.ndim == 2
        assert output.shape == (1, 1)
        assert output.index.equals(pd.RangeIndex(1))
        if col is None:
            col = pd.RangeIndex(1)
        assert output.columns.equals(pd.Index(col))
        return True

    # test scalar input
    x = 0.5

    assert is_expected_2d(d.mean())
    assert is_expected_2d(d.var())
    assert is_expected_2d(d.energy(), ["energy"])
    assert is_expected_2d(d.energy(x), ["energy"])
    assert is_expected_2d(d.pdf(x))
    assert is_expected_2d(d.log_pdf(x))
    assert is_expected_2d(d.cdf(x))
    assert is_expected_2d(d.ppf(x))
    assert is_expected_2d(d.sample())

    spl_mult = d.sample(5)
    assert isinstance(spl_mult, pd.DataFrame)
    assert spl_mult.shape == (5, 1)
    assert isinstance(spl_mult.index, pd.MultiIndex)
    assert spl_mult.index.nlevels == 2
    assert spl_mult.columns.equals(pd.RangeIndex(1))



"""Utility functions for retrieving objects from modules."""
import importlib
import inspect
import pkgutil
from functools import lru_cache

EXCLUDE_MODULES_STARTING_WITH = ("all", "test", "contrib")


def _all_functions(module_name):
    """Get all functions from a module, including submodules.

    Excludes modules starting with 'all' or 'test'.

    Parameters
    ----------
    module_name : str
        Name of the module.

    Returns
    -------
    functions_list : list
        List of tuples (function_name: str, function_object: function).
    """
    # copy to avoid modifying the cache
    return _all_cond(module_name, inspect.isfunction).copy()


def _all_classes(module_name):
    """Get all classes from a module, including submodules.

    Excludes modules starting with 'all' or 'test'.

    Parameters
    ----------
    module_name : str
        Name of the module.

    Returns
    -------
    classes_list : list
        List of tuples (class_name: str, class_ref: class).
    """
    # copy to avoid modifying the cache
    return _all_cond(module_name, inspect.isclass).copy()


@lru_cache
def _all_cond(module_name, cond):
    """Get all objects from a module satisfying a condition.

    The condition should be a hashable callable,
    of signature ``condition(obj) -> bool``.

    Excludes modules starting with 'all' or 'test'.

    Parameters
    ----------
    module_name : str
        Name of the module.
    cond : callable
        Condition to satisfy.
        Signature: ``condition(obj) -> bool``,
        passed as predicate to ``inspect.getmembers``.

    Returns
    -------
    functions_list : list
        List of tuples (function_name, function_object).
    """
    # Import the package
    package = importlib.import_module(module_name)

    # Initialize an empty list to hold all objects
    obj_list = []

    # Walk through the package's modules
    package_path = package.__path__[0]
    for _, modname, _ in pkgutil.walk_packages(
        path=[package_path], prefix=package.__name__ + "."
    ):
        # Skip modules starting with 'all' or 'test'
        if modname.split(".")[-1].startswith(EXCLUDE_MODULES_STARTING_WITH):
            continue

        # Import the module
        module = importlib.import_module(modname)

        # Get all objects from the module
        for name, obj in inspect.getmembers(module, cond):
            # if object is imported from another module, skip it
            if obj.__module__ != module.__name__:
                continue
            # add the object to the list
            obj_list.append((name, obj))

    return obj_list



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Base class for data types."""

__author__ = ["fkiraly"]

from skpro.base import BaseObject
from skpro.datatypes._common import _ret
from skpro.utils.deep_equals import deep_equals


class BaseDatatype(BaseObject):
    """Base class for data types.

    This class is the base class for all data types in sktime.
    """

    _tags = {
        "object_type": "datatype",
        "scitype": None,
        "name": None,  # any string
        "name_python": None,  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": None,
    }

    def __init__(self):
        super().__init__()

    # call defaults to check
    def __call__(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : instance of self only returned if return_metadata is True.
            Metadata dictionary.
        """
        return self._check(obj=obj, return_metadata=return_metadata, var_name=var_name)

    def check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        If self has parameters set, the check will in addition
        check whether metadata of obj is equal to self's parameters.
        In this case, ``return_metadata`` will always include the
        metadata fields required to check the parameters.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : instance of self only returned if return_metadata is True.
            Metadata dictionary.
        """
        self_params = self.get_params()

        need_check = [k for k in self_params if self_params[k] is not None]
        self_dict = {k: self_params[k] for k in need_check}

        return_metadata_orig = return_metadata

        # update return_metadata to retrieve any self_params
        # return_metadata_bool has updated condition
        if not len(need_check) == 0:
            if isinstance(return_metadata, bool):
                if not return_metadata:
                    return_metadata = need_check
                    return_metadata_bool = True
            else:
                return_metadata = set(return_metadata).union(need_check)
                return_metadata = list(return_metadata)
                return_metadata_bool = True
        elif isinstance(return_metadata, bool):
            return_metadata_bool = return_metadata
        else:
            return_metadata_bool = True

        # call inner _check
        check_res = self._check(
            obj=obj, return_metadata=return_metadata, var_name=var_name
        )

        if return_metadata_bool:
            valid = check_res[0]
            msg = check_res[1]
            metadata = check_res[2]
        else:
            valid = check_res
            msg = ""

        if not valid:
            return _ret(False, msg, None, return_metadata_orig)

        # now we know the check is valid, but we need to compare fields
        metadata_sub = {k: metadata[k] for k in self_dict}
        eqs, msg = deep_equals(self_dict, metadata_sub, return_msg=True)
        if not eqs:
            msg = f"metadata of type unequal, {msg}"
            return _ret(False, msg, None, return_metadata_orig)

        self_type = type(self)(**metadata)
        return _ret(True, "", self_type, return_metadata_orig)

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        raise NotImplementedError

    def __getitem__(self, key):
        """Get attribute by key.

        Parameters
        ----------
        key : str
            Attribute name.

        Returns
        -------
        value : any
            Attribute value.
        """
        return getattr(self, key)

    def get(self, key, default=None):
        """Get attribute by key.

        Parameters
        ----------
        key : str
            Attribute name.
        default : any, optional (default=None)
            Default value if attribute does not exist.

        Returns
        -------
        value : any
            Attribute value.
        """
        return getattr(self, key, default)

    def _get_key(self):
        """Get unique dictionary key corresponding to self.

        Private function, used in collecting a dictionary of checks.
        """
        mtype = self.get_class_tag("name")
        scitype = self.get_class_tag("scitype")
        return (mtype, scitype)


class BaseConverter(BaseObject):
    """Base class for data type converters.

    This class is the base class for all data type converters in sktime.
    """

    _tags = {
        "object_type": "converter",
        "mtype_from": None,  # type to convert from - BaseDatatype class or str
        "mtype_to": None,  # type to convert to - BaseDatatype class or str
        "multiple_conversions": False,  # whether converter encodes multiple conversions
        "python_version": None,
        "python_dependencies": None,
    }

    def __init__(self, mtype_from=None, mtype_to=None):
        self.mtype_from = mtype_from
        self.mtype_to = mtype_to
        super().__init__()

        if mtype_from is not None:
            self.set_tags(**{"mtype_from": mtype_from})
        if mtype_to is not None:
            self.set_tags(**{"mtype_to": mtype_to})

        mtype_from = self.get_tag("mtype_from")
        mtype_to = self.get_tag("mtype_to")

        if mtype_from is None:
            raise ValueError(
                f"Error in instantiating {self.__class__.__name__}: "
                "mtype_from and mtype_to must be set if the class has no defaults. "
                "For valid pairs of defaults, use get_conversions."
            )
        if mtype_to is None:
            raise ValueError(
                f"Error in instantiating {self.__class__.__name__}: "
                "mtype_to must be set in constructor, as the class has no defaults. "
                "For valid pairs of defaults, use get_conversions."
            )
        if (mtype_from, mtype_to) not in self.__class__.get_conversions():
            raise ValueError(
                f"Error in instantiating {self.__class__.__name__}: "
                "mtype_from and mtype_to must be a valid pair of defaults. "
                "For valid pairs of defaults, use get_conversions."
            )

    # call defaults to convert
    def __call__(self, obj, store=None):
        """Convert obj to another machine type.

        Parameters
        ----------
        obj : any
            Object to convert.
        store : dict, optional (default=None)
            Reference of storage for lossy conversions.

        Returns
        -------
        converted_obj : any
            Object obj converted to another machine type.
        """
        return self.convert(obj=obj, store=store)

    def convert(self, obj, store=None):
        """Convert obj to another machine type.

        Parameters
        ----------
        obj : any
            Object to convert.
        store : dict, optional (default=None)
            Reference of storage for lossy conversions.
        """
        return self._convert(obj, store)

    def _convert(self, obj, store=None):
        """Convert obj to another machine type.

        Parameters
        ----------
        obj : any
            Object to convert.
        store : dict, optional (default=None)
            Reference of storage for lossy conversions.
        """
        raise NotImplementedError

    @classmethod
    def get_conversions(cls):
        """Get all conversions.

        Returns
        -------
        list of tuples (BaseDatatype subclass, BaseDatatype subclass)
            List of all conversions in this class.
        """
        cls_from = cls.get_class_tag("mtype_from")
        cls_to = cls.get_class_tag("mtype_to")

        if cls_from is not None and cls_to is not None:
            return [(cls_from, cls_to)]
        # if multiple conversions are encoded, this should be overridden
        raise NotImplementedError

    def _get_cls_from_to(self):
        """Get classes from and to.

        Returns
        -------
        cls_from : BaseDatatype subclass
            Class to convert from.
        cls_to : BaseDatatype subclass
            Class to convert to.
        """
        cls_from = self.get_tag("mtype_from")
        cls_to = self.get_tag("mtype_to")

        cls_from = _coerce_str_to_cls(cls_from)
        cls_to = _coerce_str_to_cls(cls_to)

        return cls_from, cls_to

    def _get_key(self):
        """Get unique dictionary key corresponding to self.

        Private function, used in collecting a dictionary of checks.
        """
        cls_from, cls_to = self._get_cls_from_to()

        mtype_from = cls_from.get_class_tag("name")
        mtype_to = cls_to.get_class_tag("name")
        scitype = cls_to.get_class_tag("scitype")
        return (mtype_from, mtype_to, scitype)


class BaseExample(BaseObject):
    """Base class for Example fixtures used in tests and get_examples."""

    _tags = {
        "object_type": "datatype_example",
        "scitype": None,
        "mtype": None,
        "python_version": None,
        "python_dependencies": None,
        "index": None,  # integer index of the example to match with other mtypes
        "lossy": False,  # whether the example is lossy
    }

    def __init__(self):
        super().__init__()

    def _get_key(self):
        """Get unique dictionary key corresponding to self.

        Private function, used in collecting a dictionary of examples.
        """
        mtype = self.get_class_tag("mtype")
        scitype = self.get_class_tag("scitype")
        index = self.get_class_tag("index")
        return (mtype, scitype, index)

    def build(self):
        """Build example.

        Returns
        -------
        obj : any
            Example object.
        """
        raise NotImplementedError


def _coerce_str_to_cls(cls_or_str):
    """Get class from string.

    Parameters
    ----------
    cls_or_str : str or class
        Class or string. If string, assumed to be a unique mtype string from
        one of the BaseDatatype subclasses.

    Returns
    -------
    cls : cls_or_str, if was class; otherwise, class corresponding to string.
    """
    if not isinstance(cls_or_str, str):
        return cls_or_str

    # otherwise, we use the string to get the class from the check dict
    # perhaps it is nicer to transfer this to a registry later.
    from skpro.datatypes._check import get_check_dict

    cd = get_check_dict(soft_deps="all")
    cls = [cd[k].__class__ for k in cd if k[0] == cls_or_str]
    if len(cls) > 1:
        raise ValueError(f"Error in converting string to class: {cls_or_str}")
    elif len(cls) < 1:
        return None
    return cls[0]



"""Output coercion utilities for metric classes."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

import pandas as pd


def _coerce_to_scalar(obj):
    """Coerce obj to scalar, from polymorphic input scalar or pandas."""
    if isinstance(obj, pd.DataFrame):
        assert len(obj) == 1
        assert len(obj.columns) == 1
        return obj.iloc[0, 0]
    if isinstance(obj, pd.Series):
        assert len(obj) == 1
        return obj.iloc[0]
    return obj


def _coerce_to_df(obj):
    """Coerce to pd.DataFrame, from polymorphic input scalar or pandas."""
    return pd.DataFrame(obj)


def _coerce_to_series(obj):
    """Coerce to pd.Series, from polymorphic input scalar or pandas."""
    if isinstance(obj, pd.DataFrame):
        assert len(obj.columns) == 1
        return obj.iloc[:, 0]
    elif isinstance(obj, pd.Series):
        return obj
    else:
        return pd.Series(obj)


def _coerce_to_1d_numpy(obj):
    """Coerce to 1D np.ndarray, from pd.DataFrame or pd.Series."""
    if isinstance(obj, (pd.DataFrame, pd.Series)):
        obj = obj.values
    return obj.flatten()



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Registry and dispatcher for test classes.

Module does not contain tests, only test utilities.
"""

__author__ = ["fkiraly"]

from inspect import isclass


def get_test_class_registry():
    """Return test class registry.

    Wrapped in a function to avoid circular imports.

    Returns
    -------
    testclass_dict : dict
        test class registry
        keys are scitypes, values are test classes TestAll[Scitype]
    """
    from skpro.distributions.tests.test_all_distrs import TestAllDistributions
    from skpro.metrics.tests.test_distr_metrics import TestAllDistrMetrics
    from skpro.regression.tests.test_all_regressors import TestAllRegressors
    from skpro.tests.test_all_estimators import TestAllEstimators, TestAllObjects

    testclass_dict = dict()
    # every object in sktime inherits from BaseObject
    # "object" tests are run for all objects
    testclass_dict["object"] = TestAllObjects
    # fittable objects inherit from BaseEstimator
    # "estimator" tests are run for all estimators
    # estimators are also objects
    testclass_dict["estimator"] = TestAllEstimators
    # more specific base classes
    # these inherit either from BaseEstimator or BaseObject,
    # so also imply estimator and object tests, or only object tests
    testclass_dict["distribution"] = TestAllDistributions
    testclass_dict["regressor_proba"] = TestAllRegressors
    testclass_dict["metric_proba"] = TestAllDistrMetrics

    return testclass_dict


def get_test_classes_for_obj(obj):
    """Get all test classes relevant for an object or estimator.

    Parameters
    ----------
    obj : object or estimator, descendant of sktime BaseObject or BaseEstimator
        object or estimator for which to get test classes

    Returns
    -------
    test_classes : list of test classes
        list of test classes relevant for obj
        these are references to the actual classes, not strings
        if obj was not a descendant of BaseObject or BaseEstimator, returns empty list
    """
    from skpro.base import BaseEstimator, BaseObject
    from skpro.registry import scitype

    def is_object(obj):
        """Return whether obj is an estimator class or estimator object."""
        if isclass(obj):
            return issubclass(obj, BaseObject)
        else:
            return isinstance(obj, BaseObject)

    def is_estimator(obj):
        """Return whether obj is an estimator class or estimator object."""
        if isclass(obj):
            return issubclass(obj, BaseEstimator)
        else:
            return isinstance(obj, BaseEstimator)

    # warning: BaseEstimator does not inherit from BaseObject,
    # therefore we need to check both
    if not is_object(obj) and not is_estimator(obj):
        return []

    testclass_dict = get_test_class_registry()

    # we always need to run "object" tests
    test_clss = [testclass_dict["object"]]

    if is_estimator(obj):
        test_clss += [testclass_dict["estimator"]]

    try:
        obj_scitypes = scitype(obj, force_single_scitype=False, coerce_to_list=True)
    except Exception:
        obj_scitypes = []

    for obj_scitype in obj_scitypes:
        if obj_scitype in testclass_dict:
            test_clss += [testclass_dict[obj_scitype]]

    return test_clss



"""Adapters to sklearn linear regressors with probabilistic components."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.regression.adapters.sklearn import SklearnProbaReg
from skpro.regression.base.adapters import _DelegateWithFittedParamForwarding


class GaussianProcess(_DelegateWithFittedParamForwarding):
    """Gaussian process (GP), direct interface to sklearn GaussianProcessRegressor.

    The implementation is based on Algorithm 2.1 of [RW2006]_.

    Parameters
    ----------
    kernel : sklearn kernel instance, default=None
        The kernel specifying the covariance function of the GP. If None is
        passed, the kernel ``ConstantKernel(1.0, constant_value_bounds="fixed")
        * RBF(1.0, length_scale_bounds="fixed")`` is used as default. Note that
        the kernel hyperparameters are optimized during fitting unless the
        bounds are marked as "fixed".

    alpha : float or ndarray of shape (n_samples,), default=1e-10
        Value added to the diagonal of the kernel matrix during fitting.
        This can prevent a potential numerical issue during fitting, by
        ensuring that the calculated values form a positive definite matrix.
        It can also be interpreted as the variance of additional Gaussian
        measurement noise on the training observations. Note that this is
        different from using a `WhiteKernel`. If an array is passed, it must
        have the same number of entries as the data used for fitting and is
        used as datapoint-dependent noise level. Allowing to specify the
        noise level directly as a parameter is mainly for convenience and
        for consistency with :class:`~sklearn.linear_model.Ridge`.

    optimizer : "fmin_l_bfgs_b", callable or None, default="fmin_l_bfgs_b"
        Can either be one of the internally supported optimizers for optimizing
        the kernel's parameters, specified by a string, or an externally
        defined optimizer passed as a callable. If a callable is passed, it
        must have the signature::

            def optimizer(obj_func, initial_theta, bounds):
                # * 'obj_func': the objective function to be minimized, which
                #   takes the hyperparameters theta as a parameter and an
                #   optional flag eval_gradient, which determines if the
                #   gradient is returned additionally to the function value
                # * 'initial_theta': the initial value for theta, which can be
                #   used by local optimizers
                # * 'bounds': the bounds on the values of theta
                ....
                # Returned are the best found hyperparameters theta and
                # the corresponding value of the target function.
                return theta_opt, func_min

        Per default, the L-BFGS-B algorithm from `scipy.optimize.minimize`
        is used. If None is passed, the kernel's parameters are kept fixed.
        Available internal optimizers are: `{'fmin_l_bfgs_b'}`.

    n_restarts_optimizer : int, default=0
        The number of restarts of the optimizer for finding the kernel's
        parameters which maximize the log-marginal likelihood. The first run
        of the optimizer is performed from the kernel's initial parameters,
        the remaining ones (if any) from thetas sampled log-uniform randomly
        from the space of allowed theta-values. If greater than 0, all bounds
        must be finite. Note that `n_restarts_optimizer == 0` implies that one
        run is performed.

    normalize_y : bool, default=False
        Whether or not to normalize the target values `y` by removing the mean
        and scaling to unit-variance. This is recommended for cases where
        zero-mean, unit-variance priors are used. Note that, in this
        implementation, the normalisation is reversed before the GP predictions
        are reported.

    copy_X_train : bool, default=True
        If True, a persistent copy of the training data is stored in the
        object. Otherwise, just a reference to the training data is stored,
        which might cause predictions to change if the data is modified
        externally.

    n_targets : int, default=None
        The number of dimensions of the target values. Used to decide the number
        of outputs when sampling from the prior distributions (i.e. calling
        :meth:`sample_y` before :meth:`fit`). This parameter is ignored once
        :meth:`fit` has been called.

    random_state : int, RandomState instance or None, default=None
        Determines random number generation used to initialize the centers.
        Pass an int for reproducible results across multiple function calls.
        See :term:`Glossary <random_state>`.

    Attributes
    ----------
    kernel_ : kernel instance
        The kernel used for prediction. The structure of the kernel is the
        same as the one passed as parameter but with optimized hyperparameters.

    L_ : array-like of shape (n_samples, n_samples)
        Lower-triangular Cholesky decomposition of the kernel in ``X_train_``.

    alpha_ : array-like of shape (n_samples,)
        Dual coefficients of training data points in kernel space.

    log_marginal_likelihood_value_ : float
        The log-marginal-likelihood of ``self.kernel_.theta``.
    """

    def __init__(
        self,
        kernel=None,
        alpha=1e-10,
        optimizer="fmin_l_bfgs_b",
        n_restarts_optimizer=0,
        normalize_y=False,
        copy_X_train=True,
        n_targets=None,
        random_state=None,
    ):
        self.kernel = kernel
        self.alpha = alpha
        self.optimizer = optimizer
        self.n_restarts_optimizer = n_restarts_optimizer
        self.normalize_y = normalize_y
        self.copy_X_train = copy_X_train
        self.n_targets = n_targets
        self.random_state = random_state

        from sklearn.gaussian_process import GaussianProcessRegressor

        skl_estimator = GaussianProcessRegressor(
            kernel=kernel,
            alpha=alpha,
            optimizer=optimizer,
            n_restarts_optimizer=n_restarts_optimizer,
            normalize_y=normalize_y,
            copy_X_train=copy_X_train,
            n_targets=n_targets,
            random_state=random_state,
        )

        skpro_est = SklearnProbaReg(skl_estimator)
        self._estimator = skpro_est.clone()

        super().__init__()

    FITTED_PARAMS_TO_FORWARD = [
        "kernel_",
        "L_",
        "alpha_",
        "log_marginal_likelihood_value_",
    ]

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.gaussian_process.kernels import WhiteKernel

        param1 = {}
        param2 = {
            "kernel": WhiteKernel(),
            "alpha": 2e-10,
            "optimizer": "fmin_l_bfgs_b",
            "n_restarts_optimizer": 2,
            "normalize_y": True,
            "copy_X_train": False,
        }
        return [param1, param2]



"""Multiple quantile regression."""

__all__ = ["MultipleQuantileRegressor"]
__author__ = ["Ram0nB"]

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.base import clone

from skpro.regression.base import BaseProbaRegressor


class MultipleQuantileRegressor(BaseProbaRegressor):
    r"""Multiple quantile regressor.

    A tabular quantile regressor typically regresses a single probability level,
    however, often more than one quantile probability is of interest. Therefore, this
    multiple quantile regressor wraps multiple quantile regressors for multiple quantile
    probabilities so that each probability has one corresponding regressor. After
    fitting, all quantile regressors can be used to make probabilistic predictions.

    In `fit`, for every probability in alpha, the quantile_regressor is cloned and the
    probability is set. Subsequently all regressors are fitted.

    In probabilistic predict-like methods, if predictions of a quantile are requested,
    the fitted quantile regressor that is nearest to the desired quantile probability
    is used for the prediction, of the requested quantile.

    For instance, let :math:`\alpha = [\alpha_1, \alpha_2, \ldots, \alpha_n]`
    be the `alpha` provided to `__init__`, and
    let :math:`\alpha' = [\alpha'_1, \alpha'_2, \ldots, \alpha'_m]` be the quantiles
    requested in `predict_quantiles`.
    Then, we use the fitted quantile regressor at quantile :math:`\hat{\alpha}_j`,
    :math:`\hat{\alpha}_j := \underset{i = 1 \dots n}{\mathrm{argmin}}\ | \alpha'_j
    - \alpha_i |` to make the quantile prediction for the requested quantile
    probability :math:`\alpha'_j`.

    Consistently, the `predict_proba` method returns an empirical
    distribution with supports at quantile points corresponding to `alpha`,
    and weights corresponding to the nearest quantile regressor.

    Parameters
    ----------
    quantile_regressor : Sklearn compatible quantile regressor
        Tabular quantile regressor for probabilistic prediction methods.
    alpha_name : str, default="alpha"
        Alpha parameter name that sets the quantile probability level for the
        quantile_regressor.
    alpha : list with float, default=[0.1, 0.25, 0.5, 0.75, 0.9]
        A list of probabilities in the open interval (0, 1).
        For each probability, a quantile_regressor will be fit.
    mean_regressor : Sklearn compatible regressor, default=quantile_regressor, alpha=0.5
        Tabular mean regressor for `predict`.
    n_jobs : int or None, default=None
        The number of jobs to run in parallel for `fit` and all probabilistic prediction
        methods. -1 means using all processors, -2 means using all except one processors
        and None means no parallelization.
    sort_quantiles : bool, default=False
        For the quantile estimates, the lack of monotinicity can be a problem caused by
        the crossing of quantiles, also known as the quantile crossing problem. This can
        happen because each quantile is predicted by a single independent instance of
        the quantile_regressor. To resolve this issue, the quantiles for all predictions
        per row index will be sorted if sort_quantiles=True. This methodology will never
        lead to a higher quantile loss and solves the quantile crossing problem [1]_.

    Attributes
    ----------
    regressors_ : dict
        All fitted regressors.
        The keys are the probabilities in alpha and "mean" for the mean.

    References
    ----------
    .. [1] Victor Chernozhukov, Iván Fernández-Val, and Alfred Galichon. Quantile and
       probability curves without crossing. Econometrica, 78(3):1093-1125, 2010.

    Examples
    --------
    >>> from sklearn.linear_model import QuantileRegressor, LinearRegression
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.datasets import load_diabetes
    >>> from skpro.regression.multiquantile import MultipleQuantileRegressor
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>> alpha = [0.1, 0.3, 0.5, 0.7, 0.9]
    >>> mqr = MultipleQuantileRegressor(
    ...     mean_regressor=LinearRegression(),
    ...     quantile_regressor=QuantileRegressor(solver="highs"),
    ...     alpha_name="quantile",
    ...     alpha=alpha,
    ...     n_jobs=-1,
    ... )
    >>> mqr = mqr.fit(X_train, y_train)
    >>> predicted_quantiles = mqr.predict_quantiles(X_test, alpha)
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["ram0nb"],
        # estimator tags
        # --------------
        "capability:missing": False,
        "capability:multioutput": False,
    }

    def __init__(
        self,
        quantile_regressor=None,
        alpha_name="alpha",
        alpha=None,
        mean_regressor=None,
        n_jobs=None,
        sort_quantiles=False,
    ):
        self.quantile_regressor = quantile_regressor
        self.alpha_name = alpha_name
        self.alpha = alpha
        self.mean_regressor = mean_regressor
        self.n_jobs = n_jobs
        self.sort_quantiles = sort_quantiles

        super().__init__()

        if alpha is None:
            _alpha = [0.1, 0.25, 0.5, 0.75, 0.9]
        elif len(alpha) == 0:
            raise ValueError("at least one value in alpha is required.")
        elif np.amin(alpha) <= 0 or np.amax(alpha) >= 1:
            raise ValueError(
                "values in alpha must lie in the open interval (0, 1), "
                f"but found alpha: {alpha}."
            )
        else:
            _alpha = alpha
        self._alpha = _alpha

        if quantile_regressor is None:
            from sklearn.ensemble import GradientBoostingRegressor

            self._quantile_regressor = GradientBoostingRegressor()
        else:
            self._quantile_regressor = quantile_regressor

        if mean_regressor is None:
            self._mean_regressor = clone(self._quantile_regressor)
            self._mean_regressor.set_params(**{self.alpha_name: 0.5})
        else:
            self._mean_regressor = mean_regressor

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, Series, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        # clone, set alpha and list all regressors
        regressors = [clone(self._mean_regressor)]
        for a in self._alpha:
            q_est = clone(self._quantile_regressor)
            q_est = q_est.set_params(**{self.alpha_name: a})
            q_est_p = q_est.get_params()
            if q_est_p[self.alpha_name] != a:
                raise ValueError("Can't set alpha for quantile regressor.")
            else:
                regressors.append(q_est)

        # fit regressors
        def _fit_regressor(X, y, regressor):
            return regressor.fit(X=X, y=y)

        X_fit = X.values
        y_fit = y.values.flatten()
        fitted_regressors = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_regressor)(X_fit, y_fit, regressor) for regressor in regressors
        )

        # put fitted regressors in dict and write to self
        regressors_ = {}
        regressors_["mean"] = fitted_regressors.pop(0)
        for i, a in enumerate(self._alpha):
            regressors_[a] = fitted_regressors[i]
        self.regressors_ = regressors_

        # write varname to self ("capability:multioutput": False -> 1 column)
        self._y_varname = y.columns[0]

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`
            labels predicted for `X`
        """
        # predict
        X_pred = X.values
        preds = self.regressors_["mean"].predict(X_pred)

        # format predictions as DataFrame
        preds = np.array(preds).flatten()
        preds = pd.DataFrame(preds, columns=[self._y_varname])
        preds.index = X.index

        return preds

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        # per a in alpha, list fitted regressor that is nearest to a
        regressors = []
        for a in sorted(alpha):
            nearest_fitted_a = min(self._alpha, key=lambda p: abs(p - a))
            regressors.append(self.regressors_[nearest_fitted_a])

        # predict
        def _predict_quantile(X, regressor):
            pred = regressor.predict(X=X)
            pred = np.array(pred).flatten()
            return pred

        X_pred = X.values
        preds = Parallel(n_jobs=self.n_jobs)(
            delayed(_predict_quantile)(X_pred, regressor) for regressor in regressors
        )

        # format predictions as DataFrame
        y_cols = [self._y_varname]
        columns_sorted = pd.MultiIndex.from_product([y_cols, sorted(alpha)])
        columns = pd.MultiIndex.from_product([y_cols, alpha])
        index = X.index
        preds = pd.DataFrame(np.transpose(preds), columns=columns_sorted, index=index)

        # solve quantile crossing problem
        if self.sort_quantiles:
            preds = preds.apply(
                lambda row: row.sort_values().values, axis=1, result_type="broadcast"
            )

        # sort columns in the order of alpha requested
        preds = preds.reindex(columns=columns)

        return preds

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        from skpro.distributions import QPD_Empirical

        alpha_sorted = sorted(self._alpha)

        # get quantile prediction for all fitted quantile regressors
        quantile_preds = self._predict_quantiles(X, alpha_sorted)

        # format as emprical sample for empirical distr
        # row multiindex: (alpha, X.index)
        # column index  : as y in fit
        empirical_spl = quantile_preds.stack(level=1).swaplevel(0, 1)

        y_pred_proba = QPD_Empirical(
            empirical_spl,
            index=X.index,
            columns=[self._y_varname],
        )

        return y_pred_proba

    @classmethod
    def get_test_params(cls):
        """
        Return testing parameter settings for the regressor.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.ensemble import GradientBoostingRegressor
        from sklearn.linear_model import LinearRegression, QuantileRegressor

        from skpro.regression.tests.test_all_regressors import TEST_ALPHAS

        # extract all alpha values from TEST_ALPHAS
        alpha = []
        for a in TEST_ALPHAS:
            if isinstance(a, list):
                for a2 in a:
                    alpha.append(a2)
            else:
                alpha.append(a)

        # TEST_ALPHAS also used for predict_interval coverage, so we also need the
        # corresponding coverage quantile probabilities
        coverage_alpha = []
        for a in alpha:
            coverage_alpha.extend([0.5 - 0.5 * float(a), 0.5 + 0.5 * float(a)])
        alpha += coverage_alpha
        alpha = sorted(list(set(alpha)))

        params = [
            {
                "mean_regressor": LinearRegression(),
                "quantile_regressor": QuantileRegressor(solver="highs"),
                "alpha_name": "quantile",
                "alpha": alpha,
                "n_jobs": -1,
                "sort_quantiles": True,
            },
            {
                "mean_regressor": GradientBoostingRegressor(),
                "quantile_regressor": GradientBoostingRegressor(loss="quantile"),
                "alpha_name": "alpha",
                "alpha": alpha,
                "n_jobs": -1,
                "sort_quantiles": True,
            },
            {},  # all default values
        ]

        return params



"""Base classes for Bayesian probabilistic regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = [
    "BayesianConjugateLinearRegressor",
    "BayesianLinearRegressor",
]

from skpro.regression.bayesian._linear_conjugate import BayesianConjugateLinearRegressor
from skpro.regression.bayesian._linear_mcmc import BayesianLinearRegressor



"""Natural Gradient Boosting Regressor models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.ensemble._bagging import BaggingRegressor
from skpro.regression.ensemble._ngboost import NGBoostRegressor

__all__ = ["BaggingRegressor", "NGBoostRegressor"]
